# -*- coding: utf-8 -*-
"""
=================================================================
Mishmash generators (:mod:`sknano.nanogen._mishmash_generators`)
=================================================================

.. currentmodule:: sknano.nanogen._mishmash_generators

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

#import copy
#import itertools
#import sys

#import numpy as np

from ..chemistry import Atoms

from ._structure_generator import StructureGenerator

__all__ = ['MishmashGenerator']


class MishmashGenerator(StructureGenerator):
    u"""Class for generating a mishmash of nanostructures.

    .. note::

       This class is in development and is not functional and will
       explode if you try and use it.

    Parameters
    ----------
    autogen : bool, optional
        if `True`, automatically call
        :py:meth:`~NanotubeGenerator.generate_unit_cell`,
        followed by :py:meth:`~NanotubeGenerator.generate_structure_data`.
    verbose : bool, optional
        if `True`, show verbose output

    Examples
    --------

    """

    def __init__(self):
        super(MishmashGenerator, self).__init__()

    def generate_structure_data(self):
        """Generate structure data."""
        self._structure_atoms = Atoms()

    def save_data(self, fname=None, structure_format=None,
                  rotation_angle=None, rot_axis=None, deg2rad=True,
                  center_CM=True):
        """Save structure data.

        See :py:meth:`~sknano.nanogen.StructureGenerator.save_data` method
        for documentation.

        """
        if fname is None:
            fname = 'mishmash_structure'



# -*- coding: utf-8 -*-
"""
===========================================
Structure data files (:mod:`sknano.data`)
===========================================

.. currentmodule:: sknano.data

Contents
========

fullerenes

"""
from __future__ import division, print_function, absolute_import
__docformat__ = 'restructuredtext'

__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-
"""
=========================================================================
Helper functions for string manipulation (:mod:`sknano.tools._strfuncs`)
=========================================================================

.. currentmodule:: sknano.tools._strfuncs

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

__all__ = ['plural_word_check']


def plural_word_check(word, count):
    """Make a word plural by adding an *s* if `count` != 1.

    Parameters
    ----------
    word : str
        the word
    count : int
        the word count

    Returns
    -------
    str

    """
    return word if count == 1 else word + 's'



# -*- coding: utf-8 -*-
"""
====================================================================
Structure data atoms (:mod:`sknano.structure_io.atoms`)
====================================================================

.. currentmodule:: sknano.structure_io.atoms

.. note::

   These classes have **NOT** yet been implemented in any of the
   :py:mod:`~sknano.structure_io` classes and are still in development.
   Currently, the :py:mod:`~sknano.structure_io` classes use the
   :py:class:`~sknano.chemistry.Atom` and :py:class:`~sknano.chemistry.Atoms`
   classes provided by the :py:mod:`~sknano.chemistry` package.

Contents
========
.. autosummary::
   :toctree: generated/

   Atom
   Atoms
   LAMMPSAtom
   LAMMPSAtoms

"""
from __future__ import division, print_function, absolute_import
__docformat__ = 'restructuredtext en'

from ._atom import *
from ._atoms import *
from ._lammps_atom import *
from ._lammps_atoms import *

__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-
"""
==========================================================================
Periodic Table of Elements (:mod:`sknano.tools.refdata._periodic_table`)
==========================================================================

.. currentmodule:: sknano.tools.refdata._periodic_table

.. autosummary::
   :toctree: generated/

"""
from __future__ import division, absolute_import, print_function
__docformat__ = 'restructuredtext'

from collections import OrderedDict

from ._nist import atomic_masses_nist_list as _atomic_masses

__all__ = ['atomic_masses', 'atomic_mass_symbol_map',
           'atomic_numbers', 'atomic_number_symbol_map',
           'element_symbols', 'element_names']

element_symbols = \
    ['H', 'He',  # Period 1
     'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne',  # Period 2
     'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar',  # Period 3
     'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co',  # Period 4
     'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr',  # Period 4
     'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh',  # Period 5
     'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe',  # Period 5
     'Cs', 'Ba',  # Period 6
     'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb',  # Lanthanides
     'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu',  # Lanthanides
     'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg',  # Period 6
     'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn',  # Period 6
     'Fr', 'Ra',  # Period 7
     'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk',  # Actinides
     'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr',  # Actinides
     'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn',  # Period 7
     'Uut', 'Fl', 'Uup', 'Lv', 'Uus', 'Uuo']  # Period 7

element_names = [
    'Hydrogen', 'Helium', 'Lithium', 'Beryllium', 'Boron',
    'Carbon', 'Nitrogen', 'Oxygen', 'Fluorine', 'Neon', 'Sodium',
    'Magnesium', 'Aluminium', 'Silicon', 'Phosphorus', 'Sulfur',
    'Chlorine', 'Argon', 'Potassium', 'Calcium', 'Scandium',
    'Titanium', 'Vanadium', 'Chromium', 'Manganese', 'Iron',
    'Cobalt', 'Nickel', 'Copper', 'Zinc', 'Gallium', 'Germanium',
    'Arsenic', 'Selenium', 'Bromine', 'Krypton', 'Rubidium',
    'Strontium', 'Yttrium', 'Zirconium', 'Niobium', 'Molybdenum',
    'Technetium', 'Ruthenium', 'Rhodium', 'Palladium', 'Silver',
    'Cadmium', 'Indium', 'Tin', 'Antimony', 'Tellurium',
    'Iodine', 'Xenon', 'Caesium', 'Barium', 'Lanthanum',
    'Cerium', 'Praseodymium', 'Neodymium', 'Promethium',
    'Samarium', 'Europium', 'Gadolinium', 'Terbium',
    'Dysprosium', 'Holmium', 'Erbium', 'Thulium', 'Ytterbium',
    'Lutetium', 'Hafnium', 'Tantalum', 'Tungsten', 'Rhenium',
    'Osmium', 'Iridium', 'Platinum', 'Gold', 'Mercury',
    'Thallium', 'Lead', 'Bismuth', 'Polonium', 'Astatine',
    'Radon', 'Francium', 'Radium', 'Actinium', 'Thorium',
    'Protactinium', 'Uranium', 'Neptunium', 'Plutonium',
    'Americium', 'Curium', 'Berkelium', 'Californium',
    'Einsteinium', 'Fermium', 'Mendelevium', 'Nobelium',
    'Lawrencium', 'Rutherfordium', 'Dubnium', 'Seaborgium',
    'Bohrium', 'Hasium', 'Meitnerium', 'Darmstadtium',
    'Roentgenium', 'Copernicium', 'Ununtrium', 'Flerovium',
    'Ununpentium', 'Livermorium', 'Ununseptium', 'Ununoctium']

atomic_numbers = OrderedDict()
for Z, symbol in enumerate(element_symbols, start=1):
    atomic_numbers[symbol] = Z

atomic_number_symbol_map = {v: k for k, v in atomic_numbers.iteritems()}

atomic_masses = OrderedDict()
for i, symbol in enumerate(element_symbols):
    atomic_masses[symbol] = float(_atomic_masses[i])

atomic_mass_symbol_map = {v: k for k, v in atomic_masses.iteritems()}


class PeriodicTable(object):
    """Class for creating abstract object representing Periodic Table."""

    def __init__(self):
        pass



# -*- coding: utf-8 -*-
"""
==================================================================
Custom NumPy data structures (:mod:`sknano.tools._npcoremath`)
==================================================================

.. currentmodule:: sknano.tools._npcoremath

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import numpy as np

#from ._corefuncs import check_type

__all__ = ['NPPoint']  # , 'NPVector', 'NPQuaternion']


class NPPoint(np.ndarray):
    """Create a point in :math:`R^3`

    Parameters
    ----------
    x, y, z : float, optional
        :math:`x, y, z` coordinates of point in :math:`R^3` space.
    units : {None, str}, optional
        Units of coordinates.

    """
    def __new__(cls, x=None, y=None, z=None, units=None):
        pass
        #self._p = np.zeros(3, dtype=float)
        #self._units = units
        #for i, pi in enumerate((x, y, z)):
        #    if pi is not None:
        #        self._p[i] = pi



# -*- coding: utf-8 -*-
"""
============================================================================
scikit-nano: Python toolkit for generating and analyzing nanostructure data
============================================================================

Contents
========

Subpackages
-----------

::

 chemistry      --- abstract data structures for chemistry
 nanogen        --- modules for structure analysis and structure generation
 nanogen_gui    --- NanoGen GUI front-end
 scripts        --- command-line scripts
 structure_io   --- classes for structure data I/O
 tools          --- tools for analysis, helper funcs, LUTs, etc.

Utilitiy tools
--------------

::

 __version__    --- sknano version string

"""
from __future__ import print_function, absolute_import

__all__ = ['chemistry',
           'nanogen',
           'nanogen_gui',
           'scripts',
           'structure_io',
           'tools']

from sknano.version import version as __version__



# -*- coding: utf-8 -*-
"""
==============================================================================
Twisted nanotube bundles (:mod:`sknano.nanogen._twisted_bundle_generators`)
==============================================================================

.. currentmodule:: sknano.nanogen._twisted_bundle_generators

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

#import copy
#import itertools
#import sys

#import numpy as np

from ..tools.refdata import CCbond

from ._nanotube_bundle_generators import NanotubeBundleGenerator

__all__ = ['TwistedBundleGenerator']


class TwistedBundleGenerator(NanotubeBundleGenerator):
    u"""Class for generating twisted nanotube bundles.

    Parameters
    ----------
    n, m : int
        Chiral indices defining the nanotube chiral vector
        :math:`\\mathbf{C}_{h} = n\\mathbf{a}_{1} + m\\mathbf{a}_{2} = (n, m)`.
    nx, ny, nz : int, optional
        Number of repeat unit cells in the :math:`x, y, z` dimensions.
    element1, element2 : {str, int}, optional
        Element symbol or atomic number of basis
        :py:class:`~sknano.chemistry.Atoms` 1 and 2
    bond : float, optional
        :math:`\\mathrm{a}_{\\mathrm{CC}} =` distance between
        nearest neighbor atoms. Must be in units of **Angstroms**.
    vdw_spacing : float, optional
        van der Waals distance between nearest neighbor tubes
    bundle_packing : {None, 'hcp', 'hexagonal', 'ccp', 'cubic'}, optional
        Packing arrangement of nanotubes bundles.
        If `bundle_packing` is `None`, then it will be determined by the
        `bundle_geometry` parameter if `bundle_geometry` is not `None`.
        If both `bundle_packing` and `bundle_geometry` are `None`, then
        `bundle_packing` defaults to `hexagonal`.
    bundle_geometry : {None, 'triangle', 'hexagon', 'square', 'rectangle',
                       'rhombus', 'rhomboid'}, optional
        Force a specific geometry on the nanotube bundle boundaries.
    Lx, Ly, Lz : float, optional
        length of bundle in :math:`x, y, z` dimensions in **nanometers**.
        Overrides the :math:`n_x, n_y, n_z` cell values.
    fix_Lz : bool, optional
        Generate the nanotube with length as close to the specified
        :math:`L_z` as possible. If `True`, then
        non integer :math:`n_z` cells are permitted.
    autogen : bool, optional
        if `True`, automatically call
        :py:meth:`~NanotubeGenerator.generate_unit_cell`,
        followed by
        :py:meth:`~NanotubeBundleGenerator.generate_structure_data`.
    verbose : bool, optional
        if `True`, show verbose output

    """

    def __init__(self, n=int, m=int, nx=1, ny=1, nz=1,
                 element1='C', element2='C', bond=CCbond, vdw_spacing=3.4,
                 bundle_packing=None, bundle_geometry=None, Lx=None, Ly=None,
                 Lz=None, fix_Lz=False, autogen=True, verbose=False):

        raise RuntimeError('This class is not yet implemented.')

        super(TwistedBundleGenerator, self).__init__(
            n=n, m=m, nx=nx, ny=ny, nz=nz, element1=element1,
            element2=element2, Lx=Lx, Ly=Ly, Lz=Lz, fix_Lz=fix_Lz, bond=bond,
            autogen=False, verbose=verbose)

        if autogen:
            super(TwistedBundleGenerator, self).generate_unit_cell()
            super(TwistedBundleGenerator, self).generate_structure_data()
            self.twist_structure_data()

    def twist_structure_data(self):
        pass



# -*- coding: utf-8 -*-
"""
==============================================================
NanoGen controller (:mod:`sknano.nanogen_gui._ng_controller`)
==============================================================

.. currentmodule:: sknano.nanogen_gui._ng_controller

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from PyQt4.QtGui import QApplication

from ._ng_view import NGView

__all__ = ['NGController']


class NGController(object):
    """:mod:`~sknano.nanogen_gui` MVC controller class.

    Parameters
    ----------
    args : :attr:`python:sys.argv`
    model : :class:`~sknano.nanogen_gui._ng_model.NGModel`
        An instance of :class:`~sknano.nanogen_gui._ng_model.NGModel`.

    """
    def __init__(self, args, model=None):
        app = QApplication(args)
        self.model = model
        self.view = NGView(self, self.model)
        self.model.notify_observers()
        self.view.show()
        app.exec_()

    def refresh_view(self):
        """Refresh `NGView`."""
        self.view.update_app_view()



# -*- coding: utf-8 -*-
"""
========================================================================
Class for LAMMPS atom (:mod:`sknano.structure_io.atoms._lammps_atom`)
========================================================================

.. currentmodule:: sknano.structure_io.atoms._lammps_atom

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import numpy as np

from ...tools import check_type, Vector

from ._atom import Atom

__all__ = ['LAMMPSAtom', 'LAMMPSAtomAttributes']


class LAMMPSAtomAttributes(object):
    pass


class LAMMPSAtom(Atom):
    """`Atom` class for `LAMMPS data`.

    Parameters
    ----------
    element : {str, int}, optional
        A string representation of the element symbol or an integer specifying
        an element atomic number.
    x, y, z : float, optional
        :math:`x, y, z` components of `Atom` position vector relative to
        origin.
    atomID : int, optional
        atom ID, a LAMMPS atom attribute.
    moleculeID : int, optional
        molecule ID, a LAMMPS atom attribute.
    atomtype : int, optional
        atom type, a LAMMPS atom attribute.
    q : {int, float}, optional
        Net charge of `Atom`.
    vx, vy, vz : float, optional
        :math:`v_x, v_y, v_z` components of `Atom` velocity.
    r_units : str, optional
        Units of position components.
    v_units : str, optional
        Units of velocity components.
    nx, ny, nz : int, optional
        :math:`n_x, n_y, n_z` image flags, a LAMMPS atom attribute.
    CN : int, optional
        `Atom` coordination number.
    NN : sequence, optional
        List of nearest-neighbor `Atom` objects instances

    """

    def __init__(self, element=None, atomID=0, moleculeID=0, atomtype=1, q=0.,
                 mass=None, x=None, y=None, z=None, vx=None, vy=None, vz=None,
                 with_units=False, r_units=None, v_units=None,
                 nx=None, ny=None, nz=None, CN=None, NN=None):

        super(LAMMPSAtom, self).__init__(element=element, mass=mass,
                                         x=x, y=y, z=z)

        self._v = Vector(x=vx, y=vy, z=vz)

        self._n = np.zeros(3, dtype=int)
        for i, ni in enumerate((nx, ny, nz)):
            if ni is not None:
                self._n[i] = ni

        self._q = q
        self._atomID = int(atomID)
        self._moleculeID = int(moleculeID)
        self._atomtype = int(atomtype)
        self._CN = CN
        self._NN = NN

        self._attributes.extend(
            ['q','v', 'atomID', 'moleculeID', 'atomtype', 'CN', 'NN'])

    @property
    def CN(self):
        """Return `Atom` coordination number."""
        return self._CN

    @CN.setter
    def CN(self, value=int):
        """Set `Atom` coordination number."""
        self._CN = int(value)

    @property
    def NN(self):
        """Return list of nearest-neighbor `Atom` objects."""
        return self._NN

    @NN.setter
    def NN(self, value=list):
        """Set list of nearest-neighbor `Atom` objects."""
        self._NN = value

    @property
    def q(self):
        """Charge :math:`q` as multiple of elementary charge :math:`e`.

        """
        return self._q

    @q.setter
    def q(self, value=float):
        """Set `Atom` charge :math:`q`.

        Parameters
        ----------
        value : {int, float}
            net charge on atom as a multiple of the elementary charge
            :math:`e`.

        """
        check_type(value, allowed_types=(int, float))
        self._q = value

    @property
    def atomID(self):
        """`Atom` ID (*LAMMPS atom attribute*)."""
        return self._atomID

    @atomID.setter
    def atomID(self, value=int):
        """Set atom ID of atom.

        Parameters
        ----------
        value : int
            atom ID

        """
        check_type(value, allowed_types=(int, float))
        self._atomID = int(value)

    @property
    def moleculeID(self):
        """Molecule ID (*LAMMPS atom attribute*)."""
        return self._moleculeID

    @moleculeID.setter
    def moleculeID(self, value=int):
        """Set molecule ID of atom.

        Parameters
        ----------
        value : int
            molecule ID

        """
        check_type(value, allowed_types=(int, float))
        self._moleculeID = int(value)

    @property
    def atomtype(self):
        """`Atom` type (*LAMMPS atom attribute*)."""
        return self._atomtype

    @atomtype.setter
    def atomtype(self, value=int):
        """Set atom type of atom.

        Parameters
        ----------
        value : int
            atom type

        """
        check_type(value, allowed_types=(int, float))
        self._atomtype = int(value)

    @property
    def nx(self):
        """:math:`n_x` image flag.

        Returns
        -------
        int
            :math:`n_x` image flag.

        """
        return self._n[0]

    @nx.setter
    def nx(self, value=int):
        """Set :math:`n_x` image flag.

        Parameters
        ----------
        value : float
            :math:`n_x` image flag.

        """
        check_type(value, allowed_types=(int, float))
        self._n[0] = int(value)

    @property
    def ny(self):
        """:math:`n_y` image flag.

        Returns
        -------
        int
            :math:`n_y` image flag.

        """
        return self._n[1]

    @ny.setter
    def ny(self, value=int):
        """Set :math:`n_y` image flag.

        Parameters
        ----------
        value : float
            :math:`n_y` image flag.

        """
        check_type(value, allowed_types=(int, float))
        self._n[1] = int(value)

    @property
    def nz(self):
        """:math:`n_z` image flag.

        Returns
        -------
        int
            :math:`n_z` image flag.

        """
        return self._n[2]

    @nz.setter
    def nz(self, value=int):
        """Set `Atom` :math:`n_z` image flag.

        Parameters
        ----------
        value : float
            :math:`n_z` image flag.

        """
        check_type(value, allowed_types=(int, float))
        self._n[2] = int(value)

    @property
    def n(self):
        """:math:`n_x, n_y, n_z` image flags of `Atom`.

        Returns
        -------
        ndarray
            3-element ndarray of [:math:`n_x`, :math:`n_y`, :math:`n_z`]
            image flags of `Atom`.

        """
        return self._n

    @n.setter
    def n(self, value=np.ndarray):
        """Set :math:`n_x, n_y, n_z` image flags of `Atom`.

        Parameters
        ----------
        value : array_like
            3-element ndarray of [:math:`n_x`, :math:`n_y`, :math:`n_z`]
            image flags of `Atom`.

        """
        check_type(value, allowed_types=(np.ndarray,))
        for i, ni in enumerate(value):
            self._n[i] = ni

    @property
    def vx(self):
        """:math:`v_x` component in units of `v_units`.

        Returns
        -------
        float
            :math:`v_x` component in units of `v_units`.

        """
        return self._v.x

    @vx.setter
    def vx(self, value=float):
        """Set `Atom` :math:`v_x` component.

        Parameters
        ----------
        value : float
            :math:`v_x` component in units of `v_units`.

        """
        self._v.x = value

    @property
    def vy(self):
        """:math:`v_y` component in units of `v_units`.

        Returns
        -------
        float
            :math:`v_y` component in units of `v_units`.

        """
        return self._v.y

    @vy.setter
    def vy(self, value=float):
        """Set `Atom` :math:`v_y` component.

        Parameters
        ----------
        value : float
            :math:`v_y` component in units of `v_units`.

        """
        self._v.y = value

    @property
    def vz(self):
        """:math:`v_z` component in units of `v_units`.

        Returns
        -------
        float
            :math:`v_z` component in units of `v_units`.

        """
        return self._v.z

    @vz.setter
    def vz(self, value=float):
        """Set `Atom` :math:`v_z` component.

        Parameters
        ----------
        value : float
            :math:`v_z` component in units of `v_units`.

        """
        self._v.z = value

    @property
    def v(self):
        """:math:`v_x, v_y, v_z` velocity components in default units.

        Returns
        -------
        ndarray
            3-element ndarray of [:math:`v_x`, :math:`v_y`, :math:`v_z`]
            velocity components of `Atom`.

        """
        return self._v.components

    @v.setter
    def v(self, value=np.ndarray):
        """Set :math:`x, y, z` components of `Atom` velocity.

        Parameters
        ----------
        value : array_like
            3-element ndarray of [:math:`v_x`, :math:`v_y`, :math:`v_z`]
            velocity components of `Atom`.

        """
        self._v.components = value



# -*- coding: utf-8 -*-
"""
=============================================================
Nanotube structure tools (:mod:`sknano.nanogen._nanotubes`)
=============================================================

.. currentmodule:: sknano.nanogen._nanotubes

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from fractions import gcd
from collections import OrderedDict

import numpy as np

try:
    from pint import UnitRegistry
    ureg = UnitRegistry()
    Qty = ureg.Quantity
except ImportError:
    Qty = None

from ..chemistry import Atom
from ..tools.refdata import CCbond, dVDW

from ._parameter_luts import param_units, param_symbols, param_strfmt

__all__ = ['Nanotube', 'NanotubeBundle']


class Nanotube(object):
    u"""Class for creating interactive Nanotube objects.

    Parameters
    ----------
    n, m : int
        Chiral indices defining the nanotube chiral vector
        :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
    nx, ny, nz : int, optional
        Number of repeat unit cells in the :math:`x, y, z` dimensions
    element1, element2 : {str, int}, optional
        Element symbol or atomic number of basis
        :class:`~sknano.chemistry.Atoms` 1 and 2
    bond : float, optional
        Distance between nearest neighbor atoms (i.e., bond length).
        Must be in units of **\u212b**. Default value is
        the carbon-carbon bond length in graphite:
        :math:`\\mathrm{a}_{\\mathrm{CC}} = 1.421` \u212b ([SoFaCNTs]_)
    Lx, Ly, Lz : float, optional
        Spatial extent of nanotube(s) in :math:`x, y, z` dimensions
        in units of **nanometers**. By construction, the nanotube is oriented
        with its principle axis along the :math:`z`-axis. Therefore,
        :math:`L_z` represents the length of the nanotube(s).
        Overrides the :math:`n_x, n_y, n_z` cell values.

        .. versionadded:: 0.2.5

    tube_length : float, optional
        Length of nanotube in units of **nanometers**.
        Overrides `nz` value.

        .. deprecated:: 0.2.5
           Use `Lz` instead

    fix_Lz : bool, optional
        Generate the nanotube with length as close to the specified
        :math:`L_z` as possible. If `True`, then
        non integer :math:`n_z` cells are permitted.

        .. versionadded:: 0.2.6

    units : None, optional
        System of physical units to attach to quantities.
        **This parameter is not yet fully implemented or supported.
        Use at your own risk!**
    with_units : bool, optional
        Attach `units` to physical quantities.
        **This parameter is not yet fully implemented or supported.
        Use at your own risk!**
    verbose : bool, optional
        verbose output

    References
    ----------
    .. [SoFaCNTs] Science of Fullerenes and Carbon Nanotubes,
       M. Dresselhaus, G. Dresselhaus, P. Eklund, 1996, p. 760.

    Examples
    --------

    >>> from sknano.nanogen import Nanotube

    Create a Nanotube with :math:`\\mathbf{C}_{h} = (10, 10)` chirality.

    >>> nt = Nanotube(n=10, m=10, verbose=True)
    n: 10
    m: 10
    t₁: 1
    t₂: -1
    d: 10
    dR: 30
    N: 20
    M: 10
    R: (1, 0)
    bond: 1.421 Å
    Ch: 42.63 Å
    T: 2.46 Å
    dt: 13.57 Å
    rt: 6.78 Å
    θc: 30.00°

    Change the chirality to :math:`\\mathbf{C}_{h} = (20, 10)`.

    >>> nt.n = 20
    n: 20
    m: 10
    t₁: 4
    t₂: -5
    d: 10
    dR: 10
    N: 140
    M: 30
    R: (1, -1)
    bond: 1.421 Å
    Ch: 65.12 Å
    T: 11.28 Å
    dt: 20.73 Å
    rt: 10.36 Å
    θc: 19.11°

    Change the chirality to :math:`\\mathbf{C}_{h} = (20, 0)`.

    >>> nt.m = 0
    n: 20
    m: 0
    t₁: 1
    t₂: -2
    d: 20
    dR: 20
    N: 40
    M: 20
    R: (1, -1)
    bond: 1.421 Å
    Ch: 49.22 Å
    T: 4.26 Å
    dt: 15.67 Å
    rt: 7.83 Å
    θc: 0.00°

    """
    def __init__(self, n=int, m=int, nx=1, ny=1, nz=1, element1='C',
                 element2='C', bond=CCbond, tube_length=None,
                 Lx=None, Ly=None, Lz=None, fix_Lz=False,
                 with_units=False, units=None, verbose=False):

        if tube_length is not None and Lz is None:
            Lz = tube_length

        self._params = OrderedDict()

        # add each parameter in the order I want them to appear in
        # verbose output mode
        self._params['n'] = {}
        self._params['m'] = {}
        self._params['t1'] = {}
        self._params['t2'] = {}
        self._params['d'] = {}
        self._params['dR'] = {}
        self._params['N'] = {}
        self._params['M'] = {}
        self._params['R'] = {}
        self._params['bond'] = {}
        self._params['chiral_angle'] = {}
        self._params['Ch'] = {}
        self._params['T'] = {}
        self._params['dt'] = {}
        self._params['rt'] = {}
        self._params['electronic_type'] = {}

        self._n = int(n)
        self._m = int(m)
        self._element1 = element1
        self._element2 = element2

        if bond is None:
            bond = CCbond

        if with_units and Qty is None:
            with_units = False
        self._with_units = with_units

        if with_units and units is None:
            units = 'angstroms'
        self._units = units

        if with_units and isinstance(bond, float):
            self._bond = Qty(bond, units)
        else:
            self._bond = bond

        #self._bond = bond_lengths[element1][element2]

        self._verbose = verbose

        self._L0 = Lz  # store initial value of Lz

        self._fix_Lz = fix_Lz
        self._assume_integer_unit_cells = True
        if fix_Lz:
            self._assume_integer_unit_cells = False

        self._d = None
        self._dR = None
        self._t1 = None
        self._t2 = None
        self._N = None
        self._Ch = None
        self._T = None
        self._chiral_angle = None
        self._dt = None
        self._rt = None
        self._M = None
        self._R = None
        self._p = None
        self._q = None
        self._Lz = None
        self._electronic_type = None
        self._Natoms = None
        self._Natoms_per_tube = None

        self._Ntubes = 1
        self._nx = int(nx)
        self._ny = int(ny)

        self._Lx = Lx
        self._Ly = Ly

        if Lz is not None:
            self._T = self.compute_T(n=self._n,
                                     m=self._m,
                                     bond=self._bond,
                                     with_units=with_units,
                                     magnitude=False)
            self._Lz = float(Lz)
            if with_units:
                self._Lz = Qty(self._Lz, 'nanometers')
                self._nz = \
                    self._Lz.to('angstroms').magnitude / self._T.magnitude
            else:
                self._nz = 10 * self._Lz / self._T
        else:
            self._nz = nz

        if self._assume_integer_unit_cells:
            self._nz = int(np.ceil(self._nz))

        for k, v in self.__dict__.iteritems():
            p = k.strip('_')
            if p in self._params.keys():
                self._params[p]['units'] = param_units.get(p)
                self._params[p]['strfmt'] = param_strfmt.get(p)
                if param_symbols.get(p) is not None:
                    self._params[p]['var'] = param_symbols[p]
                else:
                    self._params[p]['var'] = p

        self.compute_tube_params()

    def compute_tube_params(self):
        """Compute/update nanotube parameters."""
        self._d = self.compute_d(n=self._n, m=self._m)
        self._dR = self.compute_dR(n=self._n, m=self._m)
        self._t1 = self.compute_t1(n=self._n, m=self._m)
        self._t2 = self.compute_t2(n=self._n, m=self._m)

        # Compute geometric/physical properties
        self._chiral_angle = self.compute_chiral_angle(n=self._n, m=self._m)
        self._Ch = self.compute_Ch(n=self._n, m=self._m, bond=self._bond,
                                   with_units=self._with_units)
        self._T = self.compute_T(n=self._n, m=self._m, bond=self._bond,
                                 with_units=self._with_units)
        self._dt = self.compute_dt(n=self._n, m=self._m, bond=self._bond,
                                   with_units=self._with_units)
        self._rt = self.compute_rt(n=self._n, m=self._m, bond=self._bond,
                                   with_units=self._with_units)

        self._M = self.compute_M(n=self._n, m=self._m)

        self._Lx = \
            self.compute_Lx(n=self._n, m=self._m, nx=self._nx, bond=self._bond,
                            with_units=self._with_units)

        self._Ly = \
            self.compute_Ly(n=self._n, m=self._m, ny=self._ny, bond=self._bond,
                            with_units=self._with_units)

        self._Lz = \
            self.compute_Lz(n=self._n, m=self._m, nz=self._nz, bond=self._bond,
                            with_units=self._with_units)
        self._tube_mass = \
            self.compute_tube_mass(n=self._n, m=self._m, nz=self._nz,
                                   element1=self._element1,
                                   element2=self._element2,
                                   with_units=self._with_units)
        self._electronic_type = \
            self.compute_electronic_type(n=self._n, m=self._m)

        # Compute symmetry properties
        self._R = self.compute_R(n=self._n, m=self._m)

        # Compute atomistic properties
        self._N = self.compute_N(n=self._n, m=self._m)
        self._Natoms = self.compute_Natoms(n=self._n, m=self._m)
        self._Natoms_per_tube = \
            self.compute_Natoms_per_tube(n=self._n, m=self._m, nz=self._nz)

        for k, v in self.__dict__.iteritems():
            p = k.strip('_')
            if p in self._params.keys():
                self._params[p]['val'] = v

        if self._verbose:
            for p, pdict in self._params.iteritems():
                pvar = pdict['var']
                try:
                    pval = pdict['val'].magnitude
                except AttributeError:
                    pval = pdict['val']
                punits = pdict['units']
                pstrfmt = pdict['strfmt']
                if punits is not None:
                    if pstrfmt is not None:
                        print(u"{}: {}{}".format(pvar,
                                                 pstrfmt.format(pval),
                                                 punits))
                    else:
                        print(u"{}: {}{}".format(pvar, pval, punits))
                else:
                    print(u"{}: {}".format(pvar, pval))
            print()

    @property
    def bond(self):
        u"""Bond length in **\u212b**."""
        return self._bond

    @bond.setter
    def bond(self, value):
        u"""Set bond length in **\u212b**."""
        self._bond = float(value)
        self.compute_tube_params()

    @property
    def element1(self):
        """Element symbol of :class:`~sknano.chemistry.Atom` 1."""
        return self._element1

    @element1.setter
    def element1(self, value):
        """Set element symbol of :class:`~sknano.chemistry.Atom` 1."""
        self._element1 = value
        self.compute_tube_params()

    @property
    def element2(self):
        """Element symbol of :class:`~sknano.chemistry.Atom` 2."""
        return self._element2

    @element2.setter
    def element2(self, value):
        """Set element symbol of :class:`~sknano.chemistry.Atom` 2."""
        self._element2 = value
        self.compute_tube_params()

    @property
    def n(self):
        """Chiral index :math:`n`.

        The component of the chiral vector :math:`\\mathbf{C}_h`
        along :math:`\\mathbf{a}_1`:

        .. math::

           \\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)

        """
        return self._n

    @n.setter
    def n(self, value):
        """Set chiral index :math:`n`"""
        self._n = int(value)
        self.compute_tube_params()

    @property
    def m(self):
        """Chiral index :math:`m`.

        The component of the chiral vector :math:`\\mathbf{C}_h`
        along :math:`\\mathbf{a}_2`:

        .. math::

           \\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)

        """
        return self._m

    @m.setter
    def m(self, value):
        """Set chiral index :math:`m`"""
        self._m = int(value)
        self.compute_tube_params()

    @property
    def d(self):
        """:math:`d=\\gcd{(n, m)}`

        :math:`d` is the **G**\ reatest **C**\ ommon **D**\ ivisor of
        :math:`n` and :math:`m`.

        """
        return self._d

    @classmethod
    def compute_d(cls, n=int, m=int):
        """Compute :math:`d=\\gcd{(n, m)}`

        :math:`d` is the **G**\ reatest **C**\ ommon **D**\ ivisor of
        :math:`n` and :math:`m`.

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.

        Returns
        -------
        int
            Greatest Common Divisor of :math:`n` and :math:`m`

        """
        return gcd(n, m)

    @property
    def dR(self):
        """:math:`d_R=\\gcd{(2n + m, 2m + n)}`

        :math:`d_R` is the **G**\ reatest **C**\ ommon **D**\ ivisor of
        :math:`2n + m` and :math:`2m + n`.

        """
        return self._dR

    @classmethod
    def compute_dR(cls, n=int, m=int):
        """Compute :math:`d_R=\\gcd{(2n + m, 2m + n)}`

        :math:`d_R` is the **G**\ reatest **C**\ ommon **D**\ ivisor of
        :math:`2n + m` and :math:`2m + n`.

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.

        Returns
        -------
        int
            greatest common divisor of :math:`2n+m` and :math:`2m+n`

        """
        return gcd(2 * m + n, 2 * n + m)

    @property
    def N(self):
        """Number of graphene hexagons in nanotube *unit cell*.

        .. math::

           N = \\frac{4(n^2 + m^2 + nm)}{d_R}

        """
        return self._N

    @classmethod
    def compute_N(cls, n=int, m=int):
        """Compute :math:`N = \\frac{2(n^2+m^2+nm)}{d_R}`.

        :math:`N` is the number of graphene hexagons mapped to a nanotube
        *unit cell*.

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.

        Returns
        -------
        int
            Number of hexagons per nanotube unit cell:
            :math:`N = \\frac{2(n^2+m^2+nm)}{d_R}`.

        """
        dR = Nanotube.compute_dR(n=n, m=m)
        return int(2 * (n**2 + m**2 + n * m) / dR)

    @property
    def t1(self):
        """:math:`t_{1} = \\frac{2m + n}{d_{R}}`

        where :math:`d_R = \\gcd{(2n + m, 2m + n)}`.

        The component of the translation vector :math:`\\mathbf{T}`
        along :math:`\\mathbf{a}_1`:

        .. math::

           \\mathbf{T} = t_1\\mathbf{a}_{1} + t_2\\mathbf{a}_2

        """
        return self._t1

    @classmethod
    def compute_t1(cls, n=int, m=int):
        """Compute :math:`t_1 = \\frac{2m + n}{d_R}`

        where :math:`d_R = \\gcd{(2n + m, 2m + n)}`.

        The component of the translation vector :math:`\\mathbf{T}`
        along :math:`\\mathbf{a}_1`:

        .. math::

           \\mathbf{T} = t_1\\mathbf{a}_{1} + t_2\\mathbf{a}_2

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.

        Returns
        -------
        int
            :math:`t_1`

        """
        dR = Nanotube.compute_dR(n=n, m=m)
        return int((2 * m + n) / dR)

    @property
    def t2(self):
        """:math:`t_2 = -\\frac{2n + m}{d_R}`

        where :math:`d_R = \\gcd{(2n + m, 2m + n)}`.

        The component of the translation vector :math:`\\mathbf{T}`
        along :math:`\\mathbf{a}_2`:

        .. math::

           \\mathbf{T} = t_1\\mathbf{a}_1 + t_2\\mathbf{a}_2

        """
        return self._t2

    @classmethod
    def compute_t2(cls, n=int, m=int):
        """Compute :math:`t_2 = -\\frac{2n + m}{d_R}`

        where :math:`d_R = \\gcd{(2n + m, 2m + n)}`.

        The component of the translation vector :math:`\\mathbf{T}`
        along :math:`\\mathbf{a}_2`:

        .. math::

           \\mathbf{T} = t_1\\mathbf{a}_1 + t_2\\mathbf{a}_2

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.

        Returns
        -------
        int
            :math:`t_2`

        """
        dR = Nanotube.compute_dR(n=n, m=m)
        return -int((2 * n + m) / dR)

    @property
    def Ch(self):
        u"""Nanotube circumference :math:`|\\mathbf{C}_h|` in **\u212b**"""
        return self._Ch

    @classmethod
    def compute_Ch(cls, n=int, m=int, bond=None, with_units=False,
                   units='angstrom', magnitude=True):
        u"""Compute nanotube circumference :math:`|\\mathbf{C}_{h}|` in
        **\u212b**.

        .. math::

           |\\mathbf{C}_h| = a\\sqrt{n^2 + m^2 + nm} =
           \\sqrt{3}a_{\\mathrm{CC}}\\sqrt{n^2 + m^2 + nm}

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        bond : float, optional
            Distance between nearest neighbor atoms.
            Must be in units of **\u212b**.

        Returns
        -------
        float
            Nanotube circumference :math:`|\\mathbf{C}_h|` in \u212b.

        """
        if bond is None:
            bond = CCbond

        if with_units and isinstance(bond, float) and Qty is not None:
            bond = Qty(bond, units)

        if magnitude and with_units:
            try:
                return bond.magnitude * np.sqrt(3 * (n**2 + m**2 + n * m))
            except AttributeError:
                return bond * np.sqrt(3 * (n**2 + m**2 + n * m))
        else:
            return bond * np.sqrt(3 * (n**2 + m**2 + n * m))

    @property
    def chiral_angle(self):
        """Chiral angle :math:`\\theta_c` in **degrees**.

        .. math::

           \\theta_c = \\tan^{-1}\\left(\\frac{\\sqrt{3} m}{2n + m}\\right)

        """
        return self._chiral_angle

    @classmethod
    def compute_chiral_angle(cls, n=int, m=int, rad2deg=True):
        """Compute chiral angle :math:`\\theta_c`.

        .. math::

           \\theta_c = \\tan^{-1}\\left(\\frac{\\sqrt{3} m}{2n + m}\\right)

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        rad2deg : bool, optional
            If `True`, return angle in degrees.

        Returns
        -------
        float
            Chiral angle :math:`\\theta_{c}` in
            degrees (default) or radians (if `rad2deg=False`).

        """
        theta = np.arctan(np.sqrt(3) * m / (2 * n + m))
        #return np.arccos((2*n + m) / (2 * np.sqrt(n**2 + m**2 + n*m)))
        if rad2deg:
            return np.degrees(theta)
        else:
            return theta

    @property
    def T(self):
        u"""Length of nanotube unit cell :math:`|\\mathbf{T}|` in \u212b.

        .. math::

           |\\mathbf{T}| = \\frac{\\sqrt{3} |\\mathbf{C}_{h}|}{d_{R}}

        """
        return self._T

    @classmethod
    def compute_T(cls, n=None, m=None, bond=None, with_units=False,
                  units='angstrom', length=True, magnitude=True):
        u"""Compute length of nanotube unit cell :math:`|\\mathbf{T}|` in
        \u212b.

        .. math::

           |\\mathbf{T}| = \\frac{\\sqrt{3} |\\mathbf{C}_{h}|}{d_{R}}

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        bond : float, optional
            Distance between nearest neighbor atoms.
            Must be in units of **\u212b**.
        units : str, optional
        with_units : bool, optional
        length : bool, optional
        magnitude : bool, optional

        Returns
        -------
        float or Qty or 2-tuple of ints
            If `length` is `True` and `with_units` is `False`, then
            return the length of unit cell in \u212b.
            If `length` is `True` and `with_units` is `True` and
            magnitude is `True`, then return the length of unit cell
            as a float, but with units in `units`.
            If `length` is `True` and `with_units` is `True` and
            magnitude is `False`, then return the length of unit cell
            as a `Quantity` instance and with units in `units`.
            If `length` is `False`, return the componets of the
            translation vector as a 2-tuple of ints (:math:`t_1`, :math:`t_2`).

        """

        if length:
            if bond is None:
                bond = CCbond

            if with_units and isinstance(bond, float) and Qty is not None:
                bond = Qty(bond, units)

            Ch = Nanotube.compute_Ch(n=n, m=m, bond=bond,
                                     with_units=with_units, units=units,
                                     magnitude=magnitude)
            dR = Nanotube.compute_dR(n=n, m=m)

            return np.sqrt(3) * Ch / dR
        else:
            t1 = Nanotube.compute_t1(n=n, m=m)
            t2 = Nanotube.compute_t2(n=n, m=m)

            return (t1, t2)

    @property
    def dt(self):
        u"""Nanotube diameter :math:`d_t = \\frac{|\\mathbf{C}_h|}{\\pi}`
        in \u212b."""
        return self._dt

    @classmethod
    def compute_dt(cls, n=int, m=int, bond=None, with_units=False,
                   units='angstrom', magnitude=True):
        u"""Compute nanotube diameter :math:`d_t` in \u212b.

        .. math::

           d_t = \\frac{|\\mathbf{C}_h|}{\\pi}

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        bond : float, optional
            Distance between nearest neighbor atoms.
            Must be in units of **\u212b**.

        Returns
        -------
        float
            Nanotube diameter :math:`d_t` in \u212b.

        """
        Ch = Nanotube.compute_Ch(n, m, bond=bond, with_units=with_units,
                                 units=units, magnitude=magnitude)
        return Ch / np.pi

    @property
    def rt(self):
        u"""Nanotube radius :math:`r_t = \\frac{|\\mathbf{C}_h|}{2\\pi}`
        in \u212b."""
        return self._rt

    @classmethod
    def compute_rt(cls, n=int, m=int, bond=None, with_units=False,
                   units='angstrom', magnitude=True):
        u"""Compute nanotube radius :math:`r_t` in \u212b.

        .. math::

           r_t = \\frac{|\\mathbf{C}_h|}{2\\pi}


        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        bond : float, optional
            Distance between nearest neighbor atoms.
            Must be in units of **\u212b**.

        Returns
        -------
        float
            Nanotube radius :math:`r_t` in \u212b.

        """
        Ch = Nanotube.compute_Ch(n=n, m=m, bond=bond, with_units=with_units,
                                 units=units, magnitude=magnitude)
        return Ch / (2 * np.pi)

    @property
    def M(self):
        """:math:`M = np - nq`

        :math:`M` is the number of multiples of the translation vector
        :math:`\\mathbf{T}` in the vector :math:`N\\mathbf{R}`.

        """
        return self._M

    @classmethod
    def compute_M(cls, n=int, m=int):
        """Compute :math:`M = mp - nq`

        :math:`M` is the number of multiples of the translation vector
        :math:`\\mathbf{T}` in the vector :math:`N\\mathbf{R}`.

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.

        Returns
        -------
        int
            :math:`M = mp - nq`

        """
        p, q = Nanotube.compute_R(n=n, m=m)
        return m * p - n * q

    @property
    def R(self):
        """Symmetry vector :math:`\\mathbf{R} = (p, q)`.

        .. math::

           \\mathbf{R} = p\\mathbf{a}_1 + q\\mathbf{a}_2

        """
        return self._R

    @classmethod
    def compute_R(cls, n=int, m=int, bond=None, with_units=False,
                  units='angstrom', length=False, magnitude=True):
        u"""Compute symmetry vector :math:`\\mathbf{R} = (p, q)`.

        The *symmetry vector* is any lattice vector of the unfolded graphene
        layer that represents a *symmetry operation* of the nanotube. The
        symmetry vector :math:`\\mathbf{R}` can be written as:

        .. math::

           \\mathbf{R} = p\\mathbf{a}_1 + q\\mathbf{a}_2

        where :math:`p` and :math:`q` are integers.
        The *symmetry vector* represents a *symmetry operation* of the nanotube
        which arises as a *screw translation*, which is a combination of
        a rotation :math:`\\psi` and translation :math:`\\tau`. The symmetry
        operation of the nanotube can be written as:

        .. math::

           R = (\\psi|\\tau)

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        length : bool, optional
            If `True`, return :math:`|\\mathbf{R}|`.
        magnitude : bool, optional
            If `True`, return :math:`|\\mathbf{R}|` in units of **\u212b**.

        Returns
        -------
        (p, q) : tuple
            2-tuple of ints -- components of :math:`\\mathbf{R}`.
        float
            Length of :math:`\\mathbf{R}` (:math:`|\\mathbf{R}|`) if `length`
            is `True` in units of **\u212b**.

        """
        t1 = Nanotube.compute_t1(n=n, m=m)
        t2 = Nanotube.compute_t2(n=n, m=m)
        N = Nanotube.compute_N(n=n, m=m)

        p = None
        q = None

        for i in xrange(0, t1 + n + 1):
            for j in xrange(t2, m + 1):
                R = t1 * j - t2 * i
                if R == 1:
                    M = m * i - n * j
                    if M > 0 and M <= N:
                        p = i
                        q = j

        if length:
            if bond is None:
                bond = CCbond

            if with_units and isinstance(bond, float) and Qty is not None:
                bond = Qty(bond, units)

            if magnitude and with_units:
                try:
                    return bond.magnitude * np.sqrt(3 * (p**2 + q**2 + p * q))
                except AttributeError:
                    return bond * np.sqrt(3 * (p**2 + q**2 + p * q))
            else:
                return bond * np.sqrt(3 * (p**2 + q**2 + p * q))
        else:
            return (p, q)

    @classmethod
    def compute_R_chiral_angle(cls, n=int, m=int, rad2deg=True):
        """Compute "chiral angle" of symmetry vector :math:`\\theta_R`.

        .. math::

           \\theta_R = \\tan^{-1}\\left(\\frac{\\sqrt{3}q}{2p + q}\\right)

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        rad2deg : bool, optional
            If `True`, return angle in degrees

        Returns
        -------
        float
            Chiral angle of *symmetry vector* :math:`\\theta_R` in
            degrees (default) or radians (if `rad2deg=False`).


        """
        p, q = Nanotube.compute_R(n=n, m=m)
        theta = np.arctan((np.sqrt(3) * q) / (2 * p + q))
        if rad2deg:
            return np.degrees(theta)
        else:
            return theta

    @classmethod
    def compute_so(cls, n=int, m=int, bond=None, with_units=False,
                   units='angstrom', magnitude=True):
        u"""Compute symmetry operation :math:`(\\psi|\\tau)`.

        The *symmetry vector* `R` represents a *symmetry
        operation* of the nanotube which arises as a *screw translation*, which
        is a combination of a rotation :math:`\\psi` and translation
        :math:`\\tau`.
        The symmetry operation of the nanotube can be written as:

        .. math::

           R = (\\psi|\\tau)

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        bond : float, optional
            Distance between nearest neighbor atoms.
            Must be in units of **\u212b**.
        rad2deg : bool, optional
            If `True`, return angle in degrees.

        Returns
        -------
        (psi, tau) : tuple
            2-tuple of floats -- :math:`\\psi` in radians and
            :math:`\\tau` in \u212b.

        """
        psi = Nanotube.compute_psi(n=n, m=m)
        tau = Nanotube.compute_tau(n=n, m=m, bond=bond, with_units=with_units,
                                   units=units, magnitude=magnitude)
        return (psi, tau)

    @classmethod
    def compute_psi(cls, n=int, m=int):
        """Compute rotation component of symmetry operation
        :math:`\\psi` in **radians**.

        .. math::

           \\psi = \\frac{2\\pi}{N}

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.

        Returns
        -------
        float
            Rotation component of symmetry operation :math:`\\psi`
            in **radians**.

        """
        N = Nanotube.compute_N(n=n, m=m)
        return 2 * np.pi / N

    @classmethod
    def compute_tau(cls, n=int, m=int, bond=None, with_units=False,
                    units='angstrom', magnitude=True):
        u"""Compute translation component of symmetry operation
        :math:`\\tau` in **\u212b**.

        .. math::

           \\tau = \\frac{M|\\mathbf{T}|}{N}

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        bond : float, optional
            Distance between nearest neighbor atoms.
            Must be in units of **\u212b**.
        rad2deg : bool, optional
            If `True`, return angle in degrees.

        Returns
        -------
        float
            Translation component of symmetry operation :math:`\\tau`
            in **\u212b**.

        """
        M = Nanotube.compute_M(n=n, m=m)
        N = Nanotube.compute_N(n=n, m=m)
        T = Nanotube.compute_T(n=n, m=m, bond=bond, with_units=with_units,
                               units=units, length=True, magnitude=magnitude)
        return M * T / N

    @property
    def nx(self):
        """Number of nanotubes along the :math:`x`-axis."""
        return int(self._nx)

    @nx.setter
    def nx(self, value=int):
        """Set :math:`n_x`"""
        self._nx = value

    @property
    def ny(self):
        """Number of nanotubes along the :math:`y`-axis."""
        return int(self._ny)

    @ny.setter
    def ny(self, value=int):
        """Set :math:`n_y`"""
        self._ny = value

    @property
    def nz(self):
        """Number of nanotube unit cells along the :math:`z`-axis."""
        return self._nz

    @nz.setter
    def nz(self, value):
        """Set number of nanotube unit cells along the :math:`z`-axis."""
        if self._assume_integer_unit_cells:
            self._nz = int(value)
        else:
            self._nz = value
        self.compute_tube_params()

    @property
    def Lx(self):
        """Spatial extent of nanotubes along :math:`x`-axis in **nm**."""
        return self._Lx

    @classmethod
    def compute_Lx(cls, n=int, m=int, nx=int, bond=None, with_units=False,
                   units='nanometer', magnitude=True):
        u"""Compute :math:`L_x` in **nanometers**.

        .. math::

           L_x = n_x (d_t + d_{\\mathrm{vdW}})

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        nx : int
            Number of nanotubes along :math:`x`-axis.
        bond : float, optional
            Bond length in **\u212b**.

        Returns
        -------
        float
            :math:`L_x` in **nanometers**

        """
        dt = Nanotube.compute_dt(n=n, m=m, bond=bond, with_units=with_units,
                                 units=units, magnitude=magnitude)
        Lx = nx * (dt + dVDW)
        if with_units:
            try:
                Lx.ito(units)
            except Exception:
                Lx.ito('nanometer')
        else:
            Lx = Lx / 10

        if magnitude and with_units:
            try:
                return Lx.magnitude
            except AttributeError:
                return Lx
        else:
            return Lx

    @property
    def Ly(self):
        """Spatial extent of nanotubes along :math:`y`-axis in **nm**."""
        return self._Ly

    @classmethod
    def compute_Ly(cls, n=int, m=int, ny=int, bond=None, with_units=False,
                   units='nanometer', magnitude=True):
        """Compute :math:`L_y` in **nanometers**.

        .. math::

           L_y = n_y (d_t + d_{\\mathrm{vdW}})

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        ny : int
            Number of nanotubes along :math:`y`-axis.
        bond : float, optional
            Bond length in **\u212b**.

        Returns
        -------
        float
            :math:`L_y` in **nanometers**

        """
        dt = Nanotube.compute_dt(n=n, m=m, bond=bond, with_units=with_units,
                                 units=units, magnitude=magnitude)
        Ly = ny * (dt + dVDW)
        if with_units:
            try:
                Ly.ito(units)
            except Exception:
                Ly.ito('nanometer')
        else:
            Ly = Ly / 10

        if magnitude and with_units:
            try:
                return Ly.magnitude
            except AttributeError:
                return Ly
        else:
            return Ly

    @property
    def Lz(self):
        """Nanotube length :math:`L_z = L_{\\mathrm{tube}}` in
        **nanometers**."""
        return self._Lz

    @Lz.setter
    def Lz(self, value=float):
        """Set tube length"""
        self._Lz = value

    @classmethod
    def compute_Lz(cls, n=int, m=int, nz=float, bond=None, with_units=False,
                   units='nanometer', magnitude=True):
        """Compute :math:`L_z = L_{\\mathrm{tube}}` in **nanometers**.

        .. math::

           L_z = n_z |\\mathbf{T}|


        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        nz : {int, float}
            Number of nanotube unit cells
        bond : float, optional
            bond length in **\u212b**.

        Returns
        -------
        float
            :math:`L_{\\mathrm{tube}}` in **nanometers**

        """
        T = Nanotube.compute_T(n=n, m=m, bond=bond, with_units=with_units,
                               units=units, length=True, magnitude=False)
        Lz = nz * T
        if with_units:
            try:
                Lz.ito(units)
            except Exception:
                Lz.ito('nanometer')
        else:
            Lz = Lz / 10

        if magnitude and with_units:
            try:
                return Lz.magnitude
            except AttributeError:
                return Lz
        else:
            return Lz

    @property
    def electronic_type(self):
        """Nanotube electronic type.

        .. versionadded:: 0.2.7

        The electronic type is determined as follows:

        if :math:`(2n + m)\\,\\mathrm{mod}\\,3=0`, the nanotube is
        **metallic**.

        if :math:`(2n + m)\\,\\mathrm{mod}\\,3=1`, the nanotube is
        **semiconducting, type 1**.

        if :math:`(2n + m)\\,\\mathrm{mod}\\,3=2`, the nanotube is
        **semiconducting, type 2**.

        The :math:`x\\,\\mathrm{mod}\\,y` notation is mathematical
        shorthand for the *modulo* operation, which computes the
        **remainder** of the division of :math:`x` by :math:`y`.
        So, for example, all *armchair* nanotubes must be metallic
        since the chiral indices satisfy: :math:`2n + m = 2n + n = 3n` and
        therefore :math:`3n\\,\\mathrm{mod}\\,3` i.e. the remainder of the
        division of :math:`3n/3=n` is always zero.

        .. note::
           Mathematically, :math:`(2n + m)\\,\\mathrm{mod}\\,3` is equivalent
           to :math:`(n - m)\\,\\mathrm{mod}\\,3` when distinguishing
           between metallic and semiconducting. However, when
           distinguishing between semiconducting types,
           one must be careful to observe the following convention:

           * Semiconducting, **type 1** means:

             * :math:`(2n + m)\\,\\mathrm{mod}\\,3=1`
             * :math:`(n - m)\\,\\mathrm{mod}\\,3=2`

           * Semiconducting, **type 2** means:

             * :math:`(2n + m)\\,\\mathrm{mod}\\,3=2`
             * :math:`(n - m)\\,\\mathrm{mod}\\,3=1`

        """
        return self._electronic_type

    @classmethod
    def compute_electronic_type(cls, n=int, m=int):
        """Compute nanotube electronic type.

        .. versionadded:: 0.2.7

        The electronic type is determined as follows:

        if :math:`(2n + m)\\,\\mathrm{mod}\\,3=0`, the nanotube is
        **metallic**.

        if :math:`(2n + m)\\,\\mathrm{mod}\\,3=1`, the nanotube is
        **semiconducting, type 1**.

        if :math:`(2n + m)\\,\\mathrm{mod}\\,3=2`, the nanotube is
        **semiconducting, type 2**.

        The :math:`x\\,\\mathrm{mod}\\,y` notation is mathematical
        shorthand for the *modulo* operation, which computes the
        **remainder** of the division of :math:`x` by :math:`y`.
        So, for example, all *armchair* nanotubes must be metallic
        since the chiral indices satisfy: :math:`2n + m = 2n + n = 3n` and
        therefore :math:`3n\\,\\mathrm{mod}\\,3` i.e. the remainder of the
        division of :math:`3n/3=n` is always zero.

        .. note::
           Mathematically, :math:`(2n + m)\\,\\mathrm{mod}\\,3` is equivalent
           to :math:`(n - m)\\,\\mathrm{mod}\\,3` when distinguishing
           between metallic and semiconducting. However, when
           distinguishing between semiconducting types,
           one must be careful to observe the following convention:

           * Semiconducting, **type 1** means:

             * :math:`(2n + m)\\,\\mathrm{mod}\\,3=1`
             * :math:`(n - m)\\,\\mathrm{mod}\\,3=2`

           * Semiconducting, **type 2** means:

             * :math:`(2n + m)\\,\\mathrm{mod}\\,3=2`
             * :math:`(n - m)\\,\\mathrm{mod}\\,3=1`

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.

        Returns
        -------
        str

        """
        if (2 * n + m) % 3 == 1:
            return 'semiconducting, type 1'
        elif (2 * n + m) % 3 == 2:
            return 'semiconducting, type 2'
        else:
            return 'metallic'

    @property
    def Natoms(self):
        """Number of atoms in nanotube *unit cell*.

        .. math::

           N_{\\mathrm{atoms}} = 2N = \\frac{4(n^2 + m^2 + nm)}{d_R}

        where :math:`N` is the number of graphene hexagons mapped to the
        nanotube unit cell.

        """
        return self._Natoms

    @classmethod
    def compute_Natoms(cls, n=int, m=int):
        """Compute :math:`N_{\mathrm{atoms/cell}} = 2N`.

        .. math::

           N_{\\mathrm{atoms}} = 2N = \\frac{4(n^2 + m^2 + nm)}{d_R}

        where :math:`N` is the number of graphene hexagons mapped to the
        nanotube unit cell.

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.

        Returns
        -------
        `2N` : int
            Number of atoms in nanotube *unit cell*:
            N_{\\mathrm{atoms}} = 2N = \\frac{4(n^2 + m^2 + nm)}{d_R}


        """
        N = Nanotube.compute_N(n=n, m=m)
        return 2 * N

    @property
    def Natoms_per_tube(self):
        """Number of atoms in nanotube :math:`N_{\\mathrm{atoms/tube}}`."""
        return self._Natoms_per_tube

    @classmethod
    def compute_Natoms_per_tube(cls, n=int, m=int, nz=float):
        """Compute :math:`N_{\\mathrm{atoms/tube}}`

        .. math::

           N_{\\mathrm{atoms/tube}} = N_{\\mathrm{atoms/cell}} \\times
           n_{z-\\mathrm{cells}}

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        nz : {int, float}
            Number of nanotube unit cells

        Returns
        -------
        int
            :math:`N_{\\mathrm{atoms/tube}}`
        """
        Natoms = Nanotube.compute_Natoms(n=n, m=m)
        return int(Natoms * nz)

    @property
    def Ntubes(self):
        """Number of nanotubes."""
        return int(self._Ntubes)

    @classmethod
    def compute_symmetry_chiral_angle(cls, n=int, m=int, rad2deg=True):
        """Alias for :meth:`Nanotube.compute_R_chiral_angle`."""
        return Nanotube.compute_R_chiral_angle(n=n, m=m, rad2deg=rad2deg)

    @classmethod
    def compute_tube_diameter(cls, n=int, m=int, bond=None, with_units=False,
                              units='angstrom', magnitude=True):
        """Alias for :meth:`Nanotube.compute_dt`"""
        return Nanotube.compute_dt(n=n, m=m, bond=bond, with_units=with_units,
                                   units=units, magnitude=magnitude)

    @classmethod
    def compute_tube_radius(cls, n=int, m=int, bond=None, with_units=False,
                            units='angstrom', magnitude=True):
        """Alias for :meth:`Nanotube.compute_rt`"""
        return Nanotube.compute_rt(n, m, bond=bond, with_units=with_units,
                                   units=units, magnitude=magnitude)

    @property
    def tube_length(self):
        """Alias for :attr:`Nanotube.Lz`"""
        return self.Lz

    @classmethod
    def compute_tube_length(cls, n=int, m=int, nz=float, bond=None,
                            with_units=False, units='nanometer',
                            magnitude=True):
        """Alias for :meth:`Nanotube.compute_Lz`"""
        return Nanotube.compute_Lz(n=n, m=m, nz=nz, bond=bond,
                                   with_units=with_units, units=units,
                                   magnitude=magnitude)

    @property
    def tube_mass(self):
        """Nanotube mass in **grams**."""
        return self._tube_mass

    @classmethod
    def compute_tube_mass(cls, n=int, m=int, nz=float, element1=None,
                          element2=None, with_units=False, units='grams',
                          magnitude=True):
        """Compute nanotube mass in **grams**.

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        nz : {int, float}
            Number of nanotube unit cells

        Returns
        -------
        float

        Notes
        -----

        .. todo::

           Handle different elements and perform accurate calculation by
           determining number of atoms of each element.

        """
        Natoms_per_tube = \
            Nanotube.compute_Natoms_per_tube(n=n, m=m, nz=nz)

        if element1 is None:
            element1 = 'C'
        if element2 is None:
            element2 = 'C'

        atom1 = Atom(element1)
        atom2 = Atom(element2)

        mass = Natoms_per_tube * (atom1.m + atom2.m) / 2
        if with_units and Qty is not None:
            mass = Qty(mass, 'Da')
            mass.ito(units)

        if magnitude and with_units:
            try:
                return mass.magnitude
            except AttributeError:
                return mass
        else:
            return mass


class NanotubeBundle(Nanotube):
    u"""Class for creating interactive Nanotube bundle objects.

    Parameters
    ----------
    n, m : int
        Chiral indices defining the nanotube chiral vector
        :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
    nx, ny, nz : int, optional
        Number of repeat unit cells in the :math:`x, y, z` dimensions
    element1, element2 : {str, int}, optional
        Element symbol or atomic number of basis
        :class:`~sknano.chemistry.Atoms` 1 and 2
    bond : float, optional
        Bond length between nearest neighbor atoms.
        Must be in units of **\u212b**. Default value is
        the carbon-carbon bond length in graphite:
        :math:`\\mathrm{a}_{\\mathrm{CC}} = 1.421` \u212b ([SoFaCNTs]_)
    Lx, Ly, Lz : float, optional
        Length of bundle in :math:`x, y, z` dimensions in **nanometers**.
        Overrides the :math:`n_x, n_y, n_z` cell values.

        .. versionadded:: 0.2.5

    tube_length : float, optional
        Length of nanotube in units of **nanometers**.

        .. deprecated:: 0.2.5
           Use `Lz` instead

    fix_Lz : bool, optional
        Generate the nanotube with length as close to the specified
        :math:`L_z` as possible. If `True`, then
        non integer :math:`n_z` cells are permitted.

        .. versionadded:: 0.2.6

    verbose : bool, optional
        Verbose output.

    """
    def __init__(self, n=int, m=int, nx=1, ny=1, nz=1,
                 element1='C', element2='C', bond=CCbond, tube_length=None,
                 vdw_spacing=dVDW, bundle_packing=None, bundle_geometry=None,
                 Lx=None, Ly=None, Lz=None, fix_Lz=False, with_units=False,
                 units=None, verbose=False):

        if tube_length is not None and Lz is None:
            Lz = tube_length

        super(NanotubeBundle, self).__init__(
            n=n, m=m, nx=nx, ny=ny, nz=nz,
            element1=element1, element2=element2,
            bond=bond, Lx=Lx, Ly=Ly, Lz=Lz, fix_Lz=fix_Lz,
            with_units=with_units, units=units, verbose=verbose)

        self._vdw_spacing = vdw_spacing
        self._bundle_packing = bundle_packing
        self._bundle_geometry = bundle_geometry

        self._Natoms_per_bundle = None

        self._bundle_mass = None
        self._bundle_density = None

        self.compute_bundle_params()

    def compute_bundle_params(self, d_vdw=None):
        """Compute/update nanotube bundle parameters."""
        super(NanotubeBundle, self).compute_tube_params()
        self._Ntubes = self.compute_Ntubes(nx=self._nx, ny=self._ny)
        self._bundle_mass = \
            self.compute_bundle_mass(n=self._n, m=self._m,
                                     nx=self._nx, ny=self._ny, nz=self._nz,
                                     with_units=False)
        self._Natoms_per_bundle = self._Ntubes * self._Natoms_per_tube
        if d_vdw is None:
            d_vdw = self._vdw_spacing

        if self._with_units and isinstance(d_vdw, float):
            d_vdw = Qty(d_vdw, 'angstrom')

        self._bundle_density = \
            self.compute_bundle_density(n=self.n, m=self.m, d_vdw=d_vdw,
                                        bond=self.bond, element1=self.element1,
                                        element2=self.element2,
                                        with_units=self._with_units)

    @property
    def Natoms_per_bundle(self):
        """Number of atoms in nanotube bundle.

        .. versionadded:: 0.2.5

        """
        return self._Natoms_per_bundle

    @Natoms_per_bundle.setter
    def Natoms_per_bundle(self, value=int):
        """Set Natoms per bundle.

        .. versionadded:: 0.2.5

        """
        self._Natoms_per_bundle = value

    @classmethod
    def compute_Natoms_per_bundle(cls, n=int, m=int, nz=float, Ntubes=None,
                                  nx=None, ny=None):
        """Compute number of atoms in nanotube bundle.

        .. versionadded:: 0.2.5

        .. math::

           N_{\\mathrm{atoms/bundle}} = N_{\\mathrm{tubes}}\\times
           N_{\\mathrm{atoms/tube}}

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        nz : {int, float}
        Ntubes : {None, int}, optional
        nx, ny : {None, int}, optional

        Returns
        -------
        Natoms_per_bundle : int
            Number of atoms in a nanotube bundle.

        Raises
        ------
        TypeError
            If `Ntubes` is `None` and `nx` or `ny` are `None`

        """
        Natoms_per_tube = \
            Nanotube.compute_Natoms_per_tube(n=n, m=m, nz=nz)
        if Ntubes is None and (nx is None or ny is None):
            raise TypeError('`Ntubes` or `nx` and `ny` must be specified as '
                            'integers')
        elif Ntubes is None and nx is not None and ny is not None:
            Ntubes = nx * ny
        Natoms_per_bundle = Ntubes * Natoms_per_tube

        return Natoms_per_bundle

    @property
    def Ntubes(self):
        """Number of nanotubes in nanotube bundle: :math:`n_x\\times n_y`."""
        return int(self._Ntubes)

    @Ntubes.setter
    def Ntubes(self, value=int):
        """Set Ntubes."""
        self._Ntubes = value
        self._nx = value
        self._ny = 1
        self.compute_bundle_params()

    @classmethod
    def compute_Ntubes(cls, nx=int, ny=int):
        """Compute number of nanotubes in nanotube bundle.

        Parameters
        ----------
        nx, ny : int, optional
            Number of repeat unit cells along :math:`x, y`-axes.

        Returns
        -------
        int
            Number of nanotubes in nanotube bundle: :math:`n_x\\times n_y`

        """
        return int(nx * ny)

    @property
    def bundle_density(self):
        """Nanotube bundle mass density :math:`\\rho_{\\mathrm{bundle}}` in
        :math:`\\mathrm{g/cm^3}`."""
        return self._bundle_density

    @classmethod
    def compute_bundle_density(cls, n=int, m=int, d_vdw=None, bond=None,
                               element1=None, element2=None,
                               with_units=False, units='g/cm**3',
                               magnitude=True):
        u"""Compute nanotube bundle mass density
        :math:`\\rho_{\\mathrm{bundle}}(n, m)` in :math:`\\mathrm{g/cm^3}`.

        .. math::

           \\rho_{\\mathrm{bundle}}(n, m) = \\frac{8\\pi^2 m_{\\mathrm{C}}
           \\sqrt{n^2 + m^2 + nm}}{9\\sqrt{3}a_{\\mathrm{CC}}^3 \\times
           \\left(\\sqrt{n^2 + m^2 + nm} +
           \\frac{\\pi d_{\\mathrm{vdW}}}{\\sqrt{3}a_{\\mathrm{CC}}}\\right)^2}

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        d_vdw : int
            van der Waals distance between nearest-neighbor tubes
        bond : float, optional
            Bond length.

        Returns
        -------
        float
            :math:`\\rho_{\\mathrm{bundle}}` in units of
            :math:`\\mathrm{\\frac{g}{cm^3}}`

        """
        if bond is None:
            bond = CCbond
        else:
            try:
                bond = bond.magnitude
            except AttributeError:
                pass

        if d_vdw is None:
            if n == m:
                d_vdw = 3.38
            elif (m == 0) or (n == 0):
                d_vdw = 3.41
            else:
                d_vdw = 3.39
        else:
            try:
                d_vdw = d_vdw.magnitude
            except AttributeError:
                pass

        if element1 is None:
            element1 = 'C'
        if element2 is None:
            element2 = 'C'

        if element1 == element2:
            bundle_density = 8 * np.pi**2 * Atom(element1).m * \
                np.sqrt(n**2 + m**2 + n*m) / \
                (9 * np.sqrt(3) * bond**3 *
                    (np.sqrt(n**2 + m**2 + n*m) +
                        np.pi * d_vdw / (np.sqrt(3) * bond))**2)
        else:
            bundle_density = 0

        if with_units and Qty is not None:
            bundle_density = Qty(bundle_density, 'Da/angstroms**3')
            bundle_density.ito(units)
            if magnitude:
                try:
                    return bundle_density.magnitude
                except AttributeError:
                    return bundle_density
            else:
                return bundle_density
        else:
            # there are 1.6605e-24 grams / Da and 1e-8 cm / angstrom
            bundle_density *= 1.6605e-24 / (1e-8)**3
            return bundle_density

    @property
    def bundle_mass(self):
        """Nanotube bundle mass :math:`M_{\\mathrm{bundle}}` in **grams**."""
        return self._bundle_mass

    @classmethod
    def compute_bundle_mass(cls, n=int, m=int, nx=int, ny=int, nz=float,
                            element1=None, element2=None, with_units=False,
                            units='grams', magnitude=True):
        """Compute nanotube bundle mass :math:`M_{\\mathrm{bundle}}`
        in **grams**.

        .. math::

           M_{\\mathrm{bundle}} = N_{\\mathrm{tubes}} M_{\\mathrm{tube}}

        Parameters
        ----------
        n, m : int
            Chiral indices defining the nanotube chiral vector
            :math:`\\mathbf{C}_h = n\\mathbf{a}_1 + m\\mathbf{a}_2 = (n, m)`.
        nx, ny, nz : int, optional
            Number of repeat unit cells in the :math:`x, y, z` dimensions
        element1, element2 : {str, int}, optional
            Element symbol or atomic number of basis
            :class:`~sknano.chemistry.Atoms` 1 and 2

        Returns
        -------
        float
            Nanotube bundle mass :math:`M_{\\mathrm{bundle}}` in **grams**.

        """
        Ntubes = \
            NanotubeBundle.compute_Ntubes(nx=nx, ny=ny)
        tube_mass = Nanotube.compute_tube_mass(n=n, m=m, nz=nz,
                                               element1=element1,
                                               element2=element2,
                                               with_units=with_units,
                                               units=units,
                                               magnitude=magnitude)
        return Ntubes * tube_mass



# -*- coding: utf-8 -*-
"""
=================================================================
Graphene structure tools (:mod:`sknano.nanogen._graphene`)
=================================================================

.. currentmodule:: sknano.nanogen._graphene

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

#import itertools

try:
    from pint import UnitRegistry
    ureg = UnitRegistry()
    Qty = ureg.Quantity
except ImportError:
    Qty = None

import numpy as np

from ..chemistry import Atom
from ..tools import Vector2D, Vector3D
from ..tools.refdata import CCbond

edge_types = {'armchair': 'AC', 'zigzag': 'ZZ'}

__all__ = ['GraphenePrimitiveCell', 'Graphene']


class GraphenePrimitiveCell(object):
    """Class for generating interactive Graphene primitive unit cell.

    .. versionadded:: 0.2.23

    Parameters
    ----------
    bond : float, optional
        bond length between nearest-neighbor atoms in **Angstroms**.

    """
    def __init__(self, bond=CCbond, with_units=False, units=None):

        if bond is None:
            bond = CCbond

        if with_units and Qty is None:
            with_units = False
        self._with_units = with_units

        if with_units and units is None:
            units = 'angstroms'
        self._units = units

        if with_units and isinstance(bond, float):
            self._bond = Qty(bond, units)
        else:
            self._bond = bond

        try:
            self._a = np.sqrt(3) * self._bond.magnitude
        except AttributeError:
            self._a = np.sqrt(3) * self._bond

        self._a1 = Vector2D(with_units=with_units, units=units)
        self._a2 = Vector2D(with_units=with_units, units=units)

        self._a1.x = self._a2.x = np.sqrt(3) / 2 * self._a
        self._a1.y = 1 / 2 * self._a
        self._a2.y = -self._a1.y

        self._b1 = Vector2D(with_units=with_units, units='1/{}'.format(units))
        self._b2 = Vector2D(with_units=with_units, units='1/{}'.format(units))

        self._b1.x = self._b2.x = 1 / np.sqrt(3) * 2 * np.pi / self._a
        self._b1.y = 2 * np.pi / self._a
        self._b2.y = -self._b1.y

    @property
    def a(self):
        """Length of graphene unit cell vector."""
        return self._a

    @property
    def a1(self):
        """:math:`a_1` unit vector."""
        return self._a1

    @property
    def a2(self):
        """:math:`a_2` unit vector."""
        return self._a2

    @property
    def b1(self):
        """:math:`b_1` reciprocal lattice vector."""
        return self._b1

    @property
    def b2(self):
        """:math:`b_2` reciprocal lattice vector."""
        return self._b2


class Graphene(object):
    """Class for generating interactive Graphene objects.

    Parameters
    ----------
    length : float, optional
        Length of graphene sheet in **nanometers**
    width : float, optional
        Width of graphene sheet in **nanometers**
    edge : {'AC', 'armchair', 'ZZ', 'zigzag'}, optional
        **A**\ rm\ **C**\ hair or **Z**\ ig\ **Z**\ ag edge along
        the `length` of the sheet.
    element1, element2 : {str, int}, optional
        Element symbol or atomic number of basis
        :class:`~sknano.chemistry.Atoms` 1 and 2
    bond : float, optional
        bond length between nearest-neighbor atoms in **Angstroms**.
    nlayers : int, optional
        Number of graphene layers.
    layer_spacing : float, optional
        Distance between layers in **Angstroms**.
    stacking_order : {'AA', 'AB'}, optional
        Stacking order of graphene layers
    autogen : bool, optional
        automatically generate unit cell and full structure
    verbose : bool, optional
        verbose output

    Notes
    -----
    For now, the graphene structure is generated using a
    conventional unit cell, not the primitive unit cell.

    .. todo::

       Add notes on unit cell calculation.

    """

    def __init__(self, length=None, width=None, edge=None,
                 element1='C', element2='C', bond=CCbond, nlayers=1,
                 layer_spacing=3.35, stacking_order='AB', with_units=False,
                 units=None, verbose=False):

        # add each parameter in the order I want them to appear in
        # verbose output mode
        self._element1 = element1
        self._element2 = element2

        self._length = length
        self._width = width
        if edge in ('armchair', 'zigzag'):
            edge = edge_types[edge]
        elif edge not in ('AC', 'ZZ'):
            print('unrecognized edge parameter: {}\n'.format(edge) +
                  'choosing one at random...\n')
            edge = np.random.choice(['AC', 'ZZ'])
            print('the randomly chosen edge type is: {}'.format(edge))
        self._edge = edge

        if bond is None:
            bond = CCbond

        if with_units and Qty is None:
            with_units = False
        self._with_units = with_units

        if with_units and units is None:
            units = 'angstroms'
        self._units = units

        if with_units and isinstance(bond, float):
            self._bond = Qty(bond, units)
        else:
            self._bond = bond

        self._verbose = verbose

        self._cell = Vector2D(with_units=with_units, units=units)

        self._Nx = 0
        self._Ny = 0

        self._layer_mass = None
        self._Natoms = 0
        self._Natoms_per_layer = None

        self._nlayers = nlayers
        self._layer_spacing = layer_spacing
        self._stacking_order = stacking_order

        self._layer_shift = Vector3D(with_units=with_units, units=units)

        if nlayers > 1 and stacking_order == 'AB':
            if edge == 'AC':
                self._layer_shift.y = self._bond
            else:
                self._layer_shift.x = self._bond

        if edge == 'AC':
            # Set up the unit cell with the armchair edge aligned
            # along the `y`-axis.
            self._cell.x = np.sqrt(3) * bond
            self._cell.y = 3 * bond
        else:
            # Set up the unit cell with the zigzag edge aligned
            # along the `y`-axis.
            self._cell.x = 3 * bond
            self._cell.y = np.sqrt(3) * bond

        self._Nx = int(np.ceil(10 * self._width / self._cell.x))
        self._Ny = int(np.ceil(10 * self._length / self._cell.y))

    def compute_layer_params(self):
        pass

    @property
    def Natoms(self):
        """Number of :class:`~sknano.chemistry.Atoms` in **unit cell**"""
        return self._Natoms

    @property
    def Natoms_per_layer(self):
        """Number of :class:`~sknano.chemistry.Atoms` in **layer**."""
        return self._Natoms_per_layer

    @classmethod
    def compute_Natoms_per_layer(cls, n=None, m=None, element1=None,
                                 element2=None, **kwargs):
        return 0

    @property
    def length(self):
        """Graphene layer length in **nanometers**."""
        return self._length

    @length.setter
    def length(self, value):
        """Set graphene layer length in **nanometers**."""
        self._length = value
        self.compute_layer_params()

    @property
    def width(self):
        """Graphene layer width in **nanometers**."""
        return self._width

    @width.setter
    def width(self, value):
        """Set graphene layer width in **nanometers**."""
        self._width = value
        self.compute_layer_params()

    @property
    def bond(self):
        """Bond length in **Angstroms**."""
        return self._bond

    @bond.setter
    def bond(self, value):
        """Set bond length in **Angstroms**."""
        self._bond = float(value)
        self.compute_layer_params()

    @property
    def edge(self):
        """Graphene edge type along length."""
        return self._edge

    @edge.setter
    def edge(self, value):
        """Set Graphene edge type along length."""
        self._edge = value
        self.compute_layer_params()

    @property
    def element1(self):
        """Element symbol of :class:`~sknano.chemistry.Atom` 1."""
        return self._element1

    @element1.setter
    def element1(self, value):
        """Set element symbol of :class:`~sknano.chemistry.Atom` 1."""
        self._element1 = value
        self.compute_layer_params()

    @property
    def element2(self):
        """Element symbol of :class:`~sknano.chemistry.Atom` 2."""
        return self._element2

    @element2.setter
    def element2(self, value):
        """Set element symbol of :class:`~sknano.chemistry.Atom` 2."""
        self._element2 = value
        self.compute_layer_params()

    @property
    def layer_mass(self):
        """Graphene layer mass in grams."""
        return self._layer_mass

    @classmethod
    def compute_layer_mass(cls, n=None, m=None, width=None, length=None,
                           edge=None, element1=None, element2=None,
                           with_units=False, units='grams', magnitude=True):
        """Compute graphene layer mass in **grams**.

        Parameters
        ----------
        n, m : int, optional
        width, length : float, optional
        edge : {'AC', 'ZZ'}
        element1, element2 : str, optional
        with_units : bool, optional
        units : str, optional
        magnitude : bool, optional

        Returns
        -------
        mass : float

        """
        Natoms_per_layer = Graphene.compute_Natoms_per_layer(
            n=n, m=m, element1=element1, element2=element2)

        if element1 is None:
            element1 = 'C'
        if element2 is None:
            element2 = 'C'

        atom1 = Atom(element1)
        atom2 = Atom(element2)

        mass = Natoms_per_layer * (atom1.m + atom2.m) / 2

        return mass

    @property
    def nlayers(self):
        return self._nlayers

    @nlayers.setter
    def nlayers(self, value):
        self._nlayers = value

    @property
    def layer_spacing(self):
        return self._layer_spacing

    @layer_spacing.setter
    def layer_spacing(self, value):
        self._layer_spacing = value



# -*- coding: utf-8 -*-
"""
==================================================
NanoGen view (:mod:`sknano.nanogen_gui._ng_view`)
==================================================

.. currentmodule:: sknano.nanogen_gui._ng_view

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

from PyQt4.QtCore import pyqtSlot
from PyQt4.QtGui import QMainWindow

from ..nanogen import NanotubeGenerator, NanotubeBundleGenerator, \
    MWNTGenerator, MWNTBundleGenerator, UnrolledNanotubeGenerator, \
    GrapheneGenerator  # , BiLayerGrapheneGenerator
from ._ui_nanogen import Ui_NanoGen

__all__ = ['NGView']


class NGView(QMainWindow, Ui_NanoGen):
    """:mod:`~sknano.nanogen_gui` MVC view class.

    Parameters
    ----------
    controller : :class:`~sknano.nanogen_gui._ng_controller.NGController`
        An instance of
        :class:`~sknano.nanogen_gui._ng_controller.NGController`.
    model : :class:`~sknano.nanogen_gui._ng_model.NGModel`
        An instance of :class:`~sknano.nanogen_gui._ng_model.NGModel`.

    """
    def __init__(self, controller=None, model=None):
        self.controller = controller
        self.model = model
        model.register_observer(self)
        super(NGView, self).__init__()
        self.setupUi(self)
        self.nanogen_statusBar.showMessage('Ready')

    # Nanotube Generator slots/signals

    @pyqtSlot()
    def on_n_spinBox_editingFinished(self):
        self.model.n = self.n_spinBox.value()

    @pyqtSlot(int)
    def on_n_spinBox_valueChanged(self, value):
        self.model.n = value

    @pyqtSlot()
    def on_m_spinBox_editingFinished(self):
        self.model.m = self.m_spinBox.value()

    @pyqtSlot(int)
    def on_m_spinBox_valueChanged(self, value):
        self.model.m = value

    @pyqtSlot()
    def on_nanotube_bond_doubleSpinBox_editingFinished(self):
        self.model.nanotube_bond = self.nanotube_bond_doubleSpinBox.value()

    @pyqtSlot(float)
    def on_nanotube_bond_doubleSpinBox_valueChanged(self, value):
        self.model.nanotube_bond = value

    @pyqtSlot()
    def on_nx_spinBox_editingFinished(self):
        self.model.nx = self.nx_spinBox.value()

    @pyqtSlot(int)
    def on_nx_spinBox_valueChanged(self, value):
        self.model.nx = value

    @pyqtSlot()
    def on_ny_spinBox_editingFinished(self):
        self.model.ny = self.ny_spinBox.value()

    @pyqtSlot(int)
    def on_ny_spinBox_valueChanged(self, value):
        self.model.ny = value

    @pyqtSlot()
    def on_nz_spinBox_editingFinished(self):
        self.model.nz = self.nz_spinBox.value()

    @pyqtSlot(int)
    def on_nz_spinBox_valueChanged(self, value):
        self.model.nz = value

    @pyqtSlot(int)
    def on_nanotube_generator_buttonGroup_buttonClicked(self):
        if self.unrolled_nanotube_generator_radioButton.isChecked():
            if self.bundle_generator_checkBox.isChecked():
                self.bundle_generator_checkBox.setCheckState(False)
            self.bundle_generator_checkBox.setCheckable(False)
        else:
            self.bundle_generator_checkBox.setCheckable(True)

    @pyqtSlot(str)
    def on_nanotube_element1_comboBox_currentIndexChanged(self, value):
        self.model.nanotube_element1 = str(value)

    @pyqtSlot(str)
    def on_nanotube_element2_comboBox_currentIndexChanged(self, value):
        self.model.nanotube_element2 = str(value)

    @pyqtSlot(str)
    def on_graphene_element1_comboBox_currentIndexChanged(self, value):
        self.model.graphene_element1 = str(value)

    @pyqtSlot(str)
    def on_graphene_element2_comboBox_currentIndexChanged(self, value):
        self.model.graphene_element2 = str(value)

    @pyqtSlot()
    def on_Lz_doubleSpinBox_editingFinished(self):
        self.model.Lz = self.Lz_doubleSpinBox.value()

    @pyqtSlot(float)
    def on_Lz_doubleSpinBox_valueChanged(self, value):
        self.model.Lz = value

    # Graphene Generator slots/signals

    #@pyqtSlot(int)
    #def on_stacking_order_buttonGroup_buttonClicked(self):
    #    if self.nlayer_spinBox.value() == 1:
    #        pass

    @pyqtSlot()
    def on_length_doubleSpinBox_editingFinished(self):
        self.model.length = self.length_doubleSpinBox.value()

    @pyqtSlot(float)
    def on_length_doubleSpinBox_valueChanged(self, value):
        self.model.length = value

    @pyqtSlot()
    def on_width_doubleSpinBox_editingFinished(self):
        self.model.width = self.width_doubleSpinBox.value()

    @pyqtSlot(float)
    def on_width_doubleSpinBox_valueChanged(self, value):
        self.model.width = value

    @pyqtSlot()
    def on_nlayers_spinBox_editingFinished(self):
        self.model.nlayers = self.nlayers_spinBox.value()
        #self.update_stacking_order_buttonGroup()

    @pyqtSlot(int)
    def on_nlayers_spinBox_valueChanged(self, value):
        self.model.nlayers = value
        #self.update_stacking_order_buttonGroup()

    #def update_stacking_order_buttonGroup(self):
    #    if self.nlayers_spinBox.value() == 1:
    #        for rb in (self.AA_stacking_radioButton,
    #                   self.AB_stacking_radioButton):
    #            rb.setCheckable(False)
    #    else:
    #        for rb in (self.AA_stacking_radioButton,
    #                   self.AB_stacking_radioButton):
    #            rb.setCheckable(True)

    @pyqtSlot()
    def on_graphene_bond_doubleSpinBox_editingFinished(self):
        self.model.graphene_bond = self.graphene_bond_doubleSpinBox.value()

    @pyqtSlot(float)
    def on_graphene_bond_doubleSpinBox_valueChanged(self, value):
        self.model.graphene_bond = value

    @pyqtSlot()
    def on_generate_pushButton_clicked(self):
        generator_tab = \
            str(self.nanogen_tabWidget.tabText(
                self.nanogen_tabWidget.currentIndex()))

        if generator_tab == 'Nanotubes':
            if self.nanotube_generator_radioButton.isChecked():
                if self.bundle_generator_checkBox.isChecked():
                    generator = NanotubeBundleGenerator(
                        n=self.n_spinBox.value(),
                        m=self.m_spinBox.value(),
                        nx=self.nx_spinBox.value(),
                        ny=self.ny_spinBox.value(),
                        nz=self.nz_spinBox.value(),
                        bond=self.nanotube_bond_doubleSpinBox.value())
                else:
                    generator = NanotubeGenerator(
                        n=self.n_spinBox.value(),
                        m=self.m_spinBox.value(),
                        nz=self.nz_spinBox.value(),
                        bond=self.nanotube_bond_doubleSpinBox.value())
            elif self.mwnt_generator_radioButton.isChecked():
                if self.bundle_generator_checkBox.isChecked():
                    generator = MWNTBundleGenerator(
                        n=self.n_spinBox.value(),
                        m=self.m_spinBox.value(),
                        nx=self.nx_spinBox.value(),
                        ny=self.ny_spinBox.value(),
                        nz=self.nz_spinBox.value(),
                        bond=self.nanotube_bond_doubleSpinBox.value())
                else:
                    generator = MWNTGenerator(
                        n=self.n_spinBox.value(),
                        m=self.m_spinBox.value(),
                        nz=self.nz_spinBox.value(),
                        bond=self.nanotube_bond_doubleSpinBox.value())
            elif self.unrolled_nanotube_generator_radioButton.isChecked():
                generator = UnrolledNanotubeGenerator(
                    n=self.n_spinBox.value(),
                    m=self.m_spinBox.value(),
                    nx=self.nx_spinBox.value(),
                    nz=self.nz_spinBox.value(),
                    bond=self.nanotube_bond_doubleSpinBox.value())
        elif generator_tab == 'Graphene':
            edge = 'AC'
            if self.ZZ_edge_radioButton.isChecked():
                edge = 'ZZ'
            stacking_order = None
            if self.AA_stacking_radioButton.isChecked():
                stacking_order = 'AA'
            elif self.AB_stacking_radioButton.isChecked():
                stacking_order = 'AB'
            generator = GrapheneGenerator(
                width=self.width_doubleSpinBox.value(),
                length=self.length_doubleSpinBox.value(),
                edge=edge,
                bond=self.graphene_bond_doubleSpinBox.value(),
                nlayers=self.nlayers_spinBox.value(),
                stacking_order=stacking_order)

        structure_format = \
            str(self.structure_format_comboBox.itemText(
                self.structure_format_comboBox.currentIndex()))
        if structure_format.endswith('data'):
            structure_format = 'data'

        generator.save_data(structure_format=structure_format)

    def update_app_view(self):
        self.n_spinBox.setValue(self.model.n)
        self.m_spinBox.setValue(self.model.m)
        self.nanotube_bond_doubleSpinBox.setValue(self.model.nanotube_bond)
        self.nx_spinBox.setValue(self.model.nx)
        self.ny_spinBox.setValue(self.model.ny)
        self.nz_spinBox.setValue(self.model.nz)
        self.Lx_lineEdit.setText('{:.4f} nm'.format(self.model.Lx))
        self.Ly_lineEdit.setText('{:.4f} nm'.format(self.model.Ly))
        self.Lz_doubleSpinBox.setValue(self.model.Lz)

        self.width_doubleSpinBox.setValue(self.model.width)
        self.length_doubleSpinBox.setValue(self.model.length)
        self.nlayers_spinBox.setValue(self.model.nlayers)
        self.graphene_bond_doubleSpinBox.setValue(self.model.graphene_bond)

        nanotube_e1e2_bond_label_text = \
            '-'.join((self.model.nanotube_element1,
                      self.model.nanotube_element2)) + ' bond ='
        self.nanotube_e1e2_bond_label.setText(nanotube_e1e2_bond_label_text)

        graphene_e1e2_bond_label_text = \
            '-'.join((self.model.graphene_element1,
                      self.model.graphene_element2)) + ' bond ='
        self.graphene_e1e2_bond_label.setText(graphene_e1e2_bond_label_text)

        #self.Ch_lineEdit.setText('{:.3f}'.format(self.model.Ch))
        #self.dt_lineEdit.setText('{:.3f}'.format(self.model.dt))
        #self.T_lineEdit.setText('{:.3f}'.format(self.model.T))
        #self.chiral_angle_lineEdit.setText(
        #    '{:.2f}'.format(self.model.chiral_angle))
        #self.N_lineEdit.setText(str(self.model.N))
        #self.Natoms_lineEdit.setText(str(self.model.Natoms))
        #self.tube_mass_lineEdit.setText(
        #    '{:.3e}'.format(self.model.tube_mass))
        #self.Natoms_per_tube_lineEdit.setText(
        #    str(self.model.Natoms_per_tube))
        #self.Ntubes_mantissa_spinBox.setValue(
        #    str(self.model.Ntubes))

        #self.d_lineEdit.setText(str(self.model.d))
        #self.dR_lineEdit.setText(str(self.model.dR))
        #self.t1_lineEdit.setText(str(self.model.t1))
        #self.t2_lineEdit.setText(str(self.model.t2))



# -*- coding: utf-8 -*-
"""
=============================================================
Parameter LUTs (:mod:`sknano.nanogen._parameter_luts`)
=============================================================

.. currentmodule:: sknano.nanogen._parameter_luts

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

param_units = {}
param_units['dt'] = \
    param_units['rt'] = \
    param_units['Ch'] = \
    param_units['T'] = \
    param_units['bond'] = u' \u212B'
param_units['chiral_angle'] = u'\u00b0'

param_symbols = {}
param_symbols['t1'] = u't\u2081'
param_symbols['t2'] = u't\u2082'
param_symbols['chiral_angle'] = u'\u03b8c'

param_strfmt = {}
param_strfmt['Ch'] = \
    param_strfmt['T'] = \
    param_strfmt['dt'] = \
    param_strfmt['rt'] = \
    param_strfmt['chiral_angle'] = '{:.2f}'
param_strfmt['bond'] = '{:.3f}'

__all__ = ['param_units', 'param_symbols', 'param_strfmt']



# -*- coding: utf-8 -*-
"""
==================================================================
Abstract representation of an Atom (:mod:`sknano.chemistry._atom`)
==================================================================

.. currentmodule:: sknano.chemistry._atom

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import numpy as np

from ..tools import check_type, Vector
from ..tools.refdata import atomic_masses, atomic_mass_symbol_map, \
    atomic_numbers, atomic_number_symbol_map, element_symbols

__all__ = ['Atom']


class Atom(object):

    """Class for creating abstract object representing an `Atom`.

    Parameters
    ----------
    element : {str, int}, optional
        A string representation of the element symbol or an integer specifying
        an element atomic number.
    atomID : int, optional
        atom ID, a LAMMPS atom attribute.
    moleculeID : int, optional
        molecule ID, a LAMMPS atom attribute.
    atomtype : int, optional
        atom type, a LAMMPS atom attribute.
    q : {int, float}, optional
        Net charge of `Atom`.
    x, y, z : float, optional
        :math:`x, y, z` components of `Atom` position vector relative to
        origin.
    vx, vy, vz : float, optional
        :math:`v_x, v_y, v_z` components of `Atom` velocity.
    with_units : bool, optional
        Attach physical units to position and velocity components.
        **Units are not fully implemented and should not be used at this
        time. This option may be removed in a future version.**
    r_units : str, optional
        Units of position components.
        **Units are not fully implemented and should not be used at this
        time. This option may be removed in a future version.**
    v_units : str, optional
        Units of velocity components.
        **Units are not fully implemented and should not be used at this
        time. This option may be removed in a future version.**
    nx, ny, nz : int, optional
        :math:`n_x, n_y, n_z` image flags, a LAMMPS atom attribute.
    CN : int, optional
        `Atom` coordination number.
    NN : sequence, optional
        List of nearest-neighbor `Atom` objects instances

    """

    def __init__(self, element=None, atomID=0, moleculeID=0, atomtype=1, q=0.,
                 mass=None, x=None, y=None, z=None, vx=None, vy=None, vz=None,
                 with_units=False, r_units=None, v_units=None,
                 nx=None, ny=None, nz=None, CN=None, NN=None):

        self._symbol = None
        self._Z = None
        self._m = None

        self._r = \
            Vector(x=x, y=y, z=z, with_units=with_units, units=r_units)

        self._v = \
            Vector(x=vx, y=vy, z=vz, with_units=with_units, units=v_units)

        self._n = np.zeros(3, dtype=int)
        for i, ni in enumerate((nx, ny, nz)):
            if ni is not None:
                self._n[i] = ni

        if isinstance(element, (int, float)):
            self._Z = int(element)
            idx = self._Z - 1
            try:
                self._symbol = element_symbols[idx]
                self._m = atomic_masses[self._symbol]
            except KeyError:
                print('unrecognized element number: {}'.format(element))
        elif isinstance(element, str):
            self._symbol = element
            try:
                self._Z = atomic_numbers[self._symbol]
                self._m = atomic_masses[self._symbol]
            except KeyError:
                print('Unrecognized atomic symbol: {}'.format(element))
        else:
            self._symbol = None
            self._Z = None
            if mass is not None and isinstance(mass, (int, float)):
                try:
                    if isinstance(mass, float):
                        self._symbol = atomic_mass_symbol_map[mass]
                    elif isinstance(mass, int):
                        self._symbol = atomic_number_symbol_map[int(mass / 2)]
                    self._Z = atomic_numbers[self._symbol]
                    self._m = atomic_masses[self._symbol]
                except KeyError:
                    self._symbol = None
                    self._Z = None
                    self._m = mass
            else:
                self._m = 0

        check_type(q, allowed_types=(int, float))
        self._q = q

        check_type(atomID, allowed_types=(int, float))
        self._atomID = int(atomID)

        check_type(moleculeID, allowed_types=(int, float))
        self._moleculeID = int(moleculeID)

        check_type(atomtype, allowed_types=(int, float))
        self._atomtype = int(atomtype)

        self._CN = CN
        self._NN = NN

        self._attributes = ['symbol', 'Z', 'm', 'q', 'r', 'v', 'atomID',
                            'moleculeID', 'atomtype', 'CN', 'NN']

    def __str__(self):
        """Return string representation of atom."""
        atom_str = ''
        for attr in self._attributes:
            atom_str += \
                'Atom {}: {}\n'.format(attr, getattr(self, '_' + attr))
        return atom_str

    @property
    def CN(self):
        """Return `Atom` coordination number."""
        return self._CN

    @CN.setter
    def CN(self, value=int):
        """Set `Atom` coordination number."""
        self._CN = int(value)

    @property
    def NN(self):
        """Return list of nearest-neighbor `Atom` objects."""
        return self._NN

    @NN.setter
    def NN(self, value=list):
        """Set list of nearest-neighbor `Atom` objects."""
        self._NN = value

    @property
    def Z(self):
        """Atomic number :math:`Z`.

        Returns
        -------
        int
            Atomic number :math:`Z`.
        """
        return self._Z

    @property
    def symbol(self):
        """Element symbol.

        Returns
        -------
        str
            Element symbol.
        """
        return self._symbol

    @property
    def m(self):
        """Atomic mass :math:`m_a` in atomic mass units.

        Returns
        -------
        float
            Atomic mass :math:`m_a` in atomic mass units.
        """
        return self._m

    @property
    def x(self):
        """:math:`x`-coordinate in units of **Angstroms**.

        Returns
        -------
        float
            :math:`x`-coordinate in units of **Angstroms**.

        """
        return self._r.x

    @x.setter
    def x(self, value=float):
        """Set `Atom` :math:`x`-coordinate in units of **Angstroms**.

        Parameters
        ----------
        value : float
            :math:`x`-coordinate in units of **Angstroms**.

        """
        self._r.x = value

    @property
    def y(self):
        """:math:`y`-coordinate in units of **Angstroms**.

        Returns
        -------
        float
            :math:`y`-coordinate in units of **Angstroms**.

        """
        return self._r.y

    @y.setter
    def y(self, value=float):
        """Set `Atom` :math:`y`-coordinate in units of **Angstroms**.

        Parameters
        ----------
        value : float
            :math:`y`-coordinate in units of **Angstroms**.

        """
        self._r.y = value

    @property
    def z(self):
        """:math:`z`-coordinate in units of **Angstroms**.

        Returns
        -------
        float
            :math:`z`-coordinate in units of **Angstroms**.

        """
        return self._r.z

    @z.setter
    def z(self, value=float):
        """Set `Atom` :math:`z`-coordinate in units of **Angstroms**.

        Parameters
        ----------
        value : float
            :math:`z`-coordinate in units of **Angstroms**.

        """
        self._r.z = value

    @property
    def r(self):
        """:math:`x, y, z` coordinates in units of **Angstroms**.

        Returns
        -------
        ndarray
            3-element ndarray of [:math:`x, y, z`] coordinates of `Atom`.

        """
        return self._r.components

    @r.setter
    def r(self, value=np.ndarray):
        """Set :math:`x, y, z` coordinates of `Atom`.

        Parameters
        ----------
        value : array_like
            3-element array of :math:`x, y, z`-coordinates in units of
            **Angstroms**.

        """
        self._r.components = value

    @property
    def q(self):
        """Charge :math:`q` as multiple of elementary charge :math:`e`.

        """
        return self._q

    @q.setter
    def q(self, value=float):
        """Set `Atom` charge :math:`q`.

        Parameters
        ----------
        value : {int, float}
            net charge on atom as a multiple of the elementary charge
            :math:`e`.

        """
        check_type(value, allowed_types=(int, float))
        self._q = value

    @property
    def atomID(self):
        """`Atom` ID (*LAMMPS atom attribute*)."""
        return self._atomID

    @atomID.setter
    def atomID(self, value=int):
        """Set atom ID of atom.

        Parameters
        ----------
        value : int
            atom ID

        """
        check_type(value, allowed_types=(int, float))
        self._atomID = int(value)

    @property
    def moleculeID(self):
        """Molecule ID (*LAMMPS atom attribute*)."""
        return self._moleculeID

    @moleculeID.setter
    def moleculeID(self, value=int):
        """Set molecule ID of atom.

        Parameters
        ----------
        value : int
            molecule ID

        """
        check_type(value, allowed_types=(int, float))
        self._moleculeID = int(value)

    @property
    def atomtype(self):
        """`Atom` type (*LAMMPS atom attribute*)."""
        return self._atomtype

    @atomtype.setter
    def atomtype(self, value=int):
        """Set atom type of atom.

        Parameters
        ----------
        value : int
            atom type

        """
        check_type(value, allowed_types=(int, float))
        self._atomtype = int(value)

    @property
    def vx(self):
        """:math:`v_x` component in units of `v_units`.

        Returns
        -------
        float
            :math:`v_x` component in units of `v_units`.

        """
        return self._v.x

    @vx.setter
    def vx(self, value=float):
        """Set `Atom` :math:`v_x` component.

        Parameters
        ----------
        value : float
            :math:`v_x` component in units of `v_units`.

        """
        self._v.x = value

    @property
    def vy(self):
        """:math:`v_y` component in units of `v_units`.

        Returns
        -------
        float
            :math:`v_y` component in units of `v_units`.

        """
        return self._v.y

    @vy.setter
    def vy(self, value=float):
        """Set `Atom` :math:`v_y` component.

        Parameters
        ----------
        value : float
            :math:`v_y` component in units of `v_units`.

        """
        self._v.y = value

    @property
    def vz(self):
        """:math:`v_z` component in units of `v_units`.

        Returns
        -------
        float
            :math:`v_z` component in units of `v_units`.

        """
        return self._v.z

    @vz.setter
    def vz(self, value=float):
        """Set `Atom` :math:`v_z` component.

        Parameters
        ----------
        value : float
            :math:`v_z` component in units of `v_units`.

        """
        self._v.z = value

    @property
    def v(self):
        """:math:`v_x, v_y, v_z` velocity components in default units.

        Returns
        -------
        ndarray
            3-element ndarray of [:math:`v_x`, :math:`v_y`, :math:`v_z`]
            velocity components of `Atom`.

        """
        return self._v.components

    @v.setter
    def v(self, value=np.ndarray):
        """Set :math:`x, y, z` components of `Atom` velocity.

        Parameters
        ----------
        value : array_like
            3-element ndarray of [:math:`v_x`, :math:`v_y`, :math:`v_z`]
            velocity components of `Atom`.

        """
        self._v.components = value

    @property
    def nx(self):
        """:math:`n_x` image flag.

        Returns
        -------
        int
            :math:`n_x` image flag.

        """
        return self._n[0]

    @nx.setter
    def nx(self, value=int):
        """Set :math:`n_x` image flag.

        Parameters
        ----------
        value : float
            :math:`n_x` image flag.

        """
        check_type(value, allowed_types=(int, float))
        self._n[0] = int(value)

    @property
    def ny(self):
        """:math:`n_y` image flag.

        Returns
        -------
        int
            :math:`n_y` image flag.

        """
        return self._n[1]

    @ny.setter
    def ny(self, value=int):
        """Set :math:`n_y` image flag.

        Parameters
        ----------
        value : float
            :math:`n_y` image flag.

        """
        check_type(value, allowed_types=(int, float))
        self._n[1] = int(value)

    @property
    def nz(self):
        """:math:`n_z` image flag.

        Returns
        -------
        int
            :math:`n_z` image flag.

        """
        return self._n[2]

    @nz.setter
    def nz(self, value=int):
        """Set `Atom` :math:`n_z` image flag.

        Parameters
        ----------
        value : float
            :math:`n_z` image flag.

        """
        check_type(value, allowed_types=(int, float))
        self._n[2] = int(value)

    @property
    def n(self):
        """:math:`n_x, n_y, n_z` image flags of `Atom`.

        Returns
        -------
        ndarray
            3-element ndarray of [:math:`n_x`, :math:`n_y`, :math:`n_z`]
            image flags of `Atom`.

        """
        return self._n

    @n.setter
    def n(self, value=np.ndarray):
        """Set :math:`n_x, n_y, n_z` image flags of `Atom`.

        Parameters
        ----------
        value : array_like
            3-element ndarray of [:math:`n_x`, :math:`n_y`, :math:`n_z`]
            image flags of `Atom`.

        """
        check_type(value, allowed_types=(np.ndarray,))
        for i, ni in enumerate(value):
            self._n[i] = ni

    def fix_minus_zero_coords(self, epsilon=1.0e-10):
        """Set really really small negative coordinates to zero.

        Set all coordinates with absolute value less than
        epsilon zero so we don't end up with -0.00000
        coordinates in structure data output.

        Parameters
        ----------
        epsilon : float
            smallest allowed absolute value of any component of atom's
            position

        """
        self._r.fix_minus_zero_components(epsilon=epsilon)

    def rezero_coords(self, epsilon=1.0e-10):
        """Re-zero position coordinates near zero.

        Set all coordinates with absolute value less than
        `epsilon` to zero.

        Parameters
        ----------
        epsilon : float
            smallest allowed absolute value of any :math:`x,y,z` component.

        """
        self._r.rezero_components(epsilon=epsilon)



# -*- coding: utf-8 -*-
"""
============================================================
ZMATRIX format (:mod:`sknano.structure_io._zmatrix_format`)
============================================================

.. currentmodule:: sknano.structure_io._zmatrix_format

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import os

from ..chemistry import Atom, Atoms
from ..tools import get_fpath

from ._structure_data import StructureReader, StructureReaderError, \
    StructureWriter, StructureConverter, default_comment_line

__all__ = ['ZMATRIXDATA', 'ZMATRIXReader', 'ZMATRIXWriter',
           'ZMATRIX2DATAConverter']


class ZMATRIXReader(StructureReader):
    """Class for reading zmatrix chemical file format.

    Parameters
    ----------
    zmatrixfile : str
        zmatrix structure file

    """
    def __init__(self, fpath=None):
        super(ZMATRIXReader, self).__init__(fpath=fpath)

        if fpath is not None:
            self.read()

    def read(self):
        with open(self.fpath, 'r') as f:
            Natoms = int(f.readline().strip())
            self._comment_line = f.readline().strip()
            lines = f.readlines()
            for line in lines:
                s = line.strip().split()
                if len(s) != 0:
                    atom = \
                        Atom(s[0], x=float(s[1]), y=float(s[2]), z=float(s[3]))
                    self._structure_atoms.append(atom)
            if self._structure_atoms.Natoms != Natoms:
                error_msg = '`zmatrixfile` contained {} atoms '.format(
                    self._structure_atoms.Natoms) + 'but should contain ' + \
                    '{}'.format(Natoms)
                raise StructureReaderError(error_msg)


class ZMATRIXWriter(StructureWriter):
    """Class for writing zmatrix chemical file format."""

    @classmethod
    def write(cls, fname=None, outpath=None, atoms=None, comment_line=None):
        """Write structure data to file.

        Parameters
        ----------
        fname : str
        outpath : str, optional
            Output path for structure data file.
        atoms : :py:class:`~sknano.chemistry.Atoms`
            An :py:class:`~sknano.chemistry.Atoms` instance.
        comment_line : str, optional
            A string written to the first line of `zmatrix` file. If `None`,
            then it is set to the full path of the output `zmatrix` file.

        """
        if not isinstance(atoms, Atoms):
            raise TypeError('atoms argument must be an `Atoms` instance')
        else:
            fpath = get_fpath(fname=fname, ext='zmatrix', outpath=outpath,
                              overwrite=True, add_fnum=False)
            if comment_line is None:
                comment_line = default_comment_line

            atoms.rezero_coords()

            with open(fpath, 'w') as f:
                f.write('{:d}\n'.format(atoms.Natoms))
                f.write('{}\n'.format(comment_line))
                for atom in atoms:
                    f.write('{:>3s}{:15.8f}{:15.8f}{:15.8f}\n'.format(
                        atom.symbol, atom.x, atom.y, atom.z))


class ZMATRIXDATA(ZMATRIXReader):
    """Class for reading and writing structure data in ZMATRIX data format.

    Parameters
    ----------
    fpath : str, optional

    """
    def __init__(self, fpath=None):
        super(ZMATRIXDATA, self).__init__(fpath=fpath)

    def write(self, zmatrixfile=None):
        """Write zmatrix file.

        Parameters
        ----------
        zmatrixfile : {None, str}, optional

        """
        try:
            if (zmatrixfile is None or zmatrixfile == '') and \
                    (self.fpath is None or self.fpath == ''):
                error_msg = '`zmatrixfile` must be a string at least 1 ' + \
                    'character long.'
                if zmatrixfile is None:
                    raise TypeError(error_msg)
                else:
                    raise ValueError(error_msg)
            else:
                zmatrixfile = self.fpath
            ZMATRIXWriter.write(fname=zmatrixfile,
                                atoms=self._structure_atoms,
                                comment_line=self._comment_line)
        except (TypeError, ValueError) as e:
            print(e)


class ZMATRIX2DATAConverter(StructureConverter):
    """
    Class for converting structure data from `zmatrix` to LAMMPS `data` format.

    Parameters
    ----------
    zmatrixfile : str

    """
    def __init__(self, zmatrixfile, boxbounds=None, pad_box=True,
                 xpad=10., ypad=10., zpad=10.):

        self._zmatrixfile = zmatrixfile
        self._datafile = os.path.splitext(self._zmatrixfile)[0] + '.data'

        super(ZMATRIX2DATAConverter, self).__init__(
            infile=self._zmatrixfile, outfile=self._datafile)

        self._boxbounds = boxbounds
        self._pad_box = pad_box
        self._xpad = xpad
        self._ypad = ypad
        self._zpad = zpad

        self._new_atoms = []
        self._add_new_atoms = False

        self._new_atomtypes = []
        self._add_new_atomtypes = False

    @property
    def zmatrixfile(self):
        return self.infile

    @property
    def datafile(self):
        """LAMMPS data file name."""
        return self.outfile

    def add_atom(self, atom=None):
        """Add new atom to atoms.

        Parameters
        ----------
        atom : `Atom`

        """
        self._new_atoms.append(atom)
        if not self._add_new_atoms:
            self._add_new_atoms = True

    def add_atomtype(self, atom=None):
        """Add new atom type to atom type dictionary.

        Parameters
        ----------
        atom : `Atom`

        """
        self._new_atomtypes.append(atom)
        if not self._add_new_atomtypes:
            self._add_new_atomtypes = True

    def convert(self, return_reader=False):
        """Convert zmatrix to LAMMPS data chemical file format.

        Parameters
        ----------
        return_reader : bool, optional
            return an instance of :py:class:`~DATAReader`

        Returns
        -------
        `DATAReader` (only if `return_reader` is True)

        """
        from ._lammps_data_format import DATAReader, DATAWriter

        zmatrixreader = ZMATRIXReader(fpath=self.infile)
        atoms = zmatrixreader.atoms
        comment_line = zmatrixreader.comment_line
        if self._add_new_atoms:
            atoms.extend(self._new_atoms)
        if self._add_new_atomtypes:
            atoms.add_atomtypes(self._new_atomtypes)

        if self._boxbounds is None:

            boxbounds = {'x': {'min': None, 'max': None},
                         'y': {'min': None, 'max': None},
                         'z': {'min': None, 'max': None}}

            for i, dim in enumerate(('x', 'y', 'z')):
                boxbounds[dim]['min'] = atoms.coords[:, i].min()
                boxbounds[dim]['max'] = atoms.coords[:, i].max()
        else:
            boxbounds = self._boxbounds

        DATAWriter.write(fname=self.outfile, atoms=atoms,
                         boxbounds=boxbounds, comment_line=comment_line,
                         pad_box=self._pad_box, xpad=self._xpad,
                         ypad=self._ypad, zpad=self._zpad)

        if return_reader:
            return DATAReader(fpath=self.outfile)



# -*- coding: utf-8 -*-
"""
====================================================================
LAMMPS dump format (:mod:`sknano.structure_io._lammps_dump_format`)
====================================================================

.. currentmodule:: sknano.structure_io._lammps_dump_format

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

import numpy as np

from ..chemistry import Atom, Atoms
from ..tools import get_fpath

from ._structure_data import StructureReader, StructureWriter, \
    StructureFormat, StructureDataError, default_comment_line


__all__ = ['DUMPReader', 'DUMPWriter', 'DUMPData', 'DUMPError', 'DUMPFormat']


class DUMPReader(StructureReader):
    """Class for reading `LAMMPS dump` file format.

    Parameters
    ----------
    fpath : str
        LAMMPS dump file path
    atom_style : {'full', 'atomic'}, optional

    """
    def __init__(self, fpath=None, atom_style='full'):
        super(DUMPReader, self).__init__(fpath=fpath)

        dump_format = DUMPFormat(atom_style=atom_style)
        self._dump_headers = dump_format.properties['headers']
        self._dump_sections = dump_format.properties['sections']
        self._section_properties = dump_format.section_properties
        self._section_syntax_dict = dump_format.section_syntax_dict

        self._headers = {}
        self._sections = {}
        self._boxbounds = {}

        if fpath is not None:
            self.read()

    @property
    def headers(self):
        """DUMP file headers."""
        return self._headers

    @property
    def sections(self):
        """DUMP file sections."""
        return self._sections

    @property
    def boxbounds(self):
        """Box bounds."""
        return self._boxbounds

    def read(self):
        """Read dump file."""
        with open(self.fpath, 'r') as f:
            self._comment_line = f.readline().strip()

            while True:
                line = f.readline().strip()
                if len(line) == 0:
                    continue
                found = False
                for key in self._dump_headers.iterkeys():
                    if key in line:
                        found = True
                        self._headers[key] = \
                            [self._dump_headers[key]['dtype'](s) for s in
                                [[ss for ss in line.split()][i] for i in
                                 range(self._dump_headers[key]['items'])]]
                        if len(self._headers[key]) == 1:
                            # since the list contains only one element,
                            # replace list value with element 0
                            self._headers[key] = self._headers[key][0]
                        break
                if not found:
                    break

            while True:
                found = False
                for section_key, header_key in self._dump_sections.iteritems():
                    if section_key in line:
                        found = True
                        f.readline()
                        Nitems = self._headers[header_key]
                        dumpdata = []
                        for n in xrange(Nitems):
                            tmp = []
                            line = f.readline().strip().split()
                            for i, props in \
                                enumerate(self._section_properties[
                                    section_key].itervalues()):
                                tmp.append(props['dtype'](line[i]))
                            dumpdata.append(tmp)
                        self._sections[section_key] = dumpdata[:]
                        break
                f.readline()
                line = f.readline().strip()
                if len(line) == 0:
                    break

        self._parse_atoms()
        self._parse_atomtypes()
        self._parse_boxbounds()

    def _parse_atoms(self):
        """Populate Atoms object with Atom objects"""
        atoms_section = self._sections['Atoms']
        atoms_section_syntax = self._section_syntax_dict['Atoms']

        masses_section = self._sections['Masses']
        masses_section_syntax = self._section_syntax_dict['Masses']

        velocities_section = self._sections['Velocities']
        velocities_section_syntax = self._section_syntax_dict['Velocities']

        atom_kwargs = {'atomID': None, 'moleculeID': None,
                       'q': None, 'atomtype': None, 'mass': None,
                       'x': None, 'y': None, 'z': None,
                       'vx': None, 'vy': None, 'vz': None}

        for lmps_atom in atoms_section:
            for kw in atom_kwargs.iterkeys():
                if kw in atoms_section_syntax:
                    atom_kwargs[kw] = \
                        lmps_atom[
                            self._section_properties['Atoms'][kw]['index']]
                elif kw in masses_section_syntax:
                    atomtype = \
                        lmps_atom[
                            self._section_properties[
                                'Atoms']['atomtype']['index']]
                    atom_kwargs[kw] = \
                        masses_section[atomtype-1][
                            self._section_properties['Masses'][kw]['index']]
                elif kw in velocities_section_syntax and \
                        len(velocities_section) == len(atoms_section):
                    atomID = \
                        lmps_atom[
                            self._section_properties[
                                'Atoms']['atomID']['index']]
                    for velocity in velocities_section:
                        velocity_atomID = \
                            velocity[self._section_properties[
                                'Velocities']['atomID']['index']]
                        if velocity_atomID == atomID:
                            atom_kwargs[kw] = \
                                velocity[self._section_properties[
                                    'Velocities'][kw]['index']]
                else:
                    print('unknown atom keyword: {}'.format(kw))

            atom = Atom(**atom_kwargs)
            self._structure_atoms.append(atom)

    def _parse_atomtypes(self):
        Ntypes = self._structure_atoms.Ntypes
        atomtypes = self._structure_atoms.atomtypes
        if Ntypes != self._headers['atom types']:
            for atomtype in xrange(1, self._headers['atom types'] + 1):
                if atomtype not in atomtypes:
                    mass = self._sections['Masses'][atomtype - 1][
                        self._section_properties['Masses']['mass']['index']]
                    self._structure_atoms.add_atomtype(
                        Atom(atomtype=atomtype, mass=mass))

    def _parse_boxbounds(self):
        for dim in ('x', 'y', 'z'):
            bounds = \
                self._headers[' '.join([dim + lim for lim in ('lo', 'hi')])]
            self._boxbounds[dim] = {'min': bounds[0], 'max': bounds[-1]}

    def get(self, section_key, colnum=None, colname=None, colindex=None):
        """Return section with `section_key`.

        Parameters
        ----------
        section_key : str
        colnum : int, optional
        colname : str, optional
        colindex : int, optional

        Returns
        -------
        sequence

        """
        section_data = None
        try:
            section_data = self._sections[section_key]
            section_syntax = self._section_syntax_dict[section_key]
        except KeyError as e:
            print(e)
        else:
            try:
                colidx = None
                if colnum is not None:
                    colidx = int(colnum - 1)
                elif colname is not None:
                    colidx = \
                        self._section_properties[section_key][colname]['index']
                elif colindex is not None:
                    colidx = int(colindex)
            except (KeyError, TypeError, ValueError) as e:
                print(e)
            else:
                try:
                    colname = section_syntax[colidx]
                    coltype = \
                        self._section_properties[section_key][colname]['dtype']
                    section_data = \
                        np.asarray(
                            section_data, dtype=coltype)[:, colidx].tolist()
                except TypeError:
                    pass
        finally:
            return section_data


class DUMPWriter(StructureWriter):
    """Class for writing LAMMPS dump chemical file format."""

    @classmethod
    def write(cls, fname=None, outpath=None, atoms=None, boxbounds=None,
              comment_line=None, assume_unique_atoms=False, verbose=False):
        """Write structure dump to file.

        Parameters
        ----------
        fname : str
        outpath : str, optional
            Output path for structure data file.
        atoms : `Atoms`
            An :py:class:`Atoms` instance.
        boxbounds : dict, optional
            If `None`, determined automatically from atom coordinates.
        comment_line : str, optional
            A string written to the first line of `dump` file. If `None`,
            then it is set to the full path of the output `dump` file.
        assume_unique_atoms : bool, optional
            Check that each Atom in Atoms has a unique atomID. If the check
            fails, then assign a unique atomID to each Atom.
            If `assume_unique_atoms` is True, but the atomID's are not
            unique, LAMMPS will not be able to read the dump file.
        verbose : bool, optional
            verbose output

        """
        if not isinstance(atoms, Atoms):
            raise TypeError('atoms argument must be an `Atoms` instance')
        else:
            fpath = get_fpath(fname=fname, ext='dump', outpath=outpath,
                              overwrite=True, add_fnum=False)
            if comment_line is None:
                comment_line = default_comment_line

            atoms.rezero_coords()

            atomtypes = atoms.atomtypes

            Natoms = atoms.Natoms
            Natoms_width = \
                8 if len(str(Natoms)) <= 12 else len(str(Natoms)) + 4
            Ntypes = atoms.Ntypes
            Ntypes_width = Natoms_width

            atomID_width = len(str(Natoms)) + 1
            atomtype_width = len(str(Ntypes)) + 1

            if not assume_unique_atoms and \
                    len(set(atoms.atom_ids)) != atoms.Natoms:
                for atomID, atom in enumerate(atoms, start=1):
                    atom.atomID = atomID

            if boxbounds is None:
                boxbounds = {'x': {'min': None, 'max': None},
                             'y': {'min': None, 'max': None},
                             'z': {'min': None, 'max': None}}

                for i, dim in enumerate(('x', 'y', 'z')):
                    boxbounds[dim]['min'] = atoms.coords[:, i].min()
                    boxbounds[dim]['max'] = atoms.coords[:, i].max()

            lohi_width = 0
            for dim in ('x', 'y', 'z'):
                lohi_width = \
                    max(lohi_width, len('{:.6f} {:.6f}'.format(
                        boxbounds[dim]['min'], boxbounds[dim]['max'])) + 4)

            with open(fpath, 'w') as f:
                f.write('# {}\n\n'.format(comment_line.lstrip('#').strip()))
                f.write('{}atoms\n'.format(
                    '{:d}'.format(Natoms).ljust(Natoms_width)))
                f.write('{}atom types\n\n'.format(
                    '{:d}'.format(Ntypes).ljust(Ntypes_width)))

                for dim in ('x', 'y', 'z'):
                    f.write('{}{dim}lo {dim}hi\n'.format(
                        '{:.6f} {:.6f}'.format(
                            boxbounds[dim]['min'],
                            boxbounds[dim]['max']).ljust(lohi_width),
                        dim=dim))

                f.write('\nMasses\n\n')
                for atomtype, properties in atomtypes.iteritems():
                    f.write('{}{:.4f}\n'.format(
                        '{:d}'.format(atomtype).ljust(Natoms_width),
                        properties['mass']))

                f.write('\nAtoms\n\n')
                for atom in atoms:
                    line = ''
                    line += "{:>{}}".format(atom.atomID, atomID_width)
                    line += "{:>{}}".format(atom.moleculeID, 3)
                    line += "{:>{}}".format(atom.atomtype, atomtype_width)
                    line += "{:>{}}".format('{:.1f}'.format(atom.q), 4)
                    line += "{:>{}}".format('{:f}'.format(atom.x), 14)
                    line += "{:>{}}".format('{:f}'.format(atom.y), 14)
                    line += "{:>{}}".format('{:f}'.format(atom.z), 14)
                    line += "{:>{}}".format('{:d}'.format(atom.nx), 3)
                    line += "{:>{}}".format('{:d}'.format(atom.ny), 3)
                    line += "{:>{}}\n".format('{:d}'.format(atom.nz), 3)

                    f.write(line)

                f.write('\nVelocities\n\n')
                for atom in atoms:
                    line = ''
                    line += "{:>{}}".format(atom.atomID, atomID_width)
                    line += "{:>{}}".format('{:f}'.format(atom.vx), 14)
                    line += "{:>{}}".format('{:f}'.format(atom.vy), 14)
                    line += "{:>{}}\n".format('{:f}'.format(atom.vz), 14)

                    f.write(line)


class aselect(object):
    pass


class tselect(object):
    pass


class DUMPData(DUMPReader):
    """Class for reading and writing structure data in LAMMPS dump format.

    Parameters
    ----------
    fpath : str, optional

    """
    def __init__(self, fpath=None):
        super(DUMPData, self).__init__(fpath=fpath)
        self._snaps = []
        self._nsnaps = self._nselect = 0
        self._names = {}
        self._aselect = aselect(self)
        self._tselect = tselect(self)

    def atom(self, n, l=[]):
        pass

    def read_all(self):
        pass

    def next(self):
        pass

    def read_snapshot(self, f):
        pass

    def scale(self, *list):
        pass

    def scale_one(self, snap, x, y, z):
        pass

    def unscale(self, *list):
        pass

    def unscale_one(self, snap, x, y, z):
        pass

    def wrap(self):
        pass

    def unwrap(self):
        pass

    def owrap(self):
        pass

    def names2str(self):
        pass

    def sort(self, *list):
        pass

    def sort_one(self, snap, id):
        pass

    def scatter(self, root):
        pass

    def minmax(self, colname):
        pass

    def set(self, eq):
        pass

    def setv(self, colname, vec):
        pass

    def clone(self, nstep, col):
        pass

    def spread(self, old, n, new):
        pass

    def time(self):
        pass

    def vecs(self, n, *list):
        pass

    def newcolumn(self, str):
        pass

    def compare_atom(self, a, b):
        pass

    def compare_time(self, a, b):
        pass

    def cull(self):
        pass

    def delete(self):
        pass

    def extra(self):
        pass

    def findtime(self, n):
        pass

    def map(self, *pairs):
        pass

    def maxbox(self):
        pass

    def maxtype(self):
        pass

    def replace(self, section_key, new_data, colnum=None,
                colname=None, colindex=None):
        """Replace section data.

        Parameters
        ----------
        section_key : str
        new_data : sequence
        colnum : int, optional
        colname : str, optional
        colindex : int, optional

        """
        colidx = None

        # for backwards compatibility with the pizza.py dump module,
        # first check positional arguments to see if this method was called
        # using the pizza.py dump module signature which expects positional
        # arguments of the following type:
        #   data.replace(str, int, list)
        if isinstance(new_data, (int, float)) and \
                isinstance(colnum, (np.ndarray, list)):
            colidx = int(new_data) - 1
            new_data = np.asarray(colnum)
        else:
            try:
                colidx = None
                if colnum is not None:
                    colidx = int(colnum - 1)
                elif colname is not None:
                    colidx = \
                        self._section_properties[section_key][colname]['index']
                elif colindex is not None:
                    colidx = int(colindex)
            except (KeyError, TypeError, ValueError) as e:
                print(e)
                raise DUMPError('replace called with invalid arguments')
        atom_attr = self._section_syntax_dict[section_key][colidx]
        attr_dtype = \
            self._section_properties[section_key][atom_attr]['dtype']
        new_data = np.asarray(new_data, dtype=attr_dtype)

        for i, atom in enumerate(self._structure_atoms):
            self._sections[section_key][i][colidx] = attr_dtype(new_data[i])
            setattr(atom, atom_attr, attr_dtype(new_data[i]))

    def write(self, dumpfile=None):
        """Write dump file.

        Parameters
        ----------
        dumpfile : {None, str}, optional

        """
        try:
            if (dumpfile is None or dumpfile == '') and \
                    (self.fpath is None or self.fpath == ''):
                error_msg = '`dumpfile` must be a string at least 1 ' + \
                    'character long.'
                if dumpfile is None:
                    raise TypeError(error_msg)
                else:
                    raise ValueError(error_msg)
            else:
                dumpfile = self.fpath
            DUMPWriter.write(fname=dumpfile, atoms=self._structure_atoms,
                             boxbounds=self._boxbounds,
                             comment_line=self._comment_line)
        except (TypeError, ValueError) as e:
            print(e)


class DUMPError(StructureDataError):
    """Exception raised for failed method calls.

    Parameters
    ----------
    msg : str
        Error message

    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class DUMPFormat(StructureFormat):
    """Class defining the structure file format for LAMMPS dump."""
    pass



# -*- coding: utf-8 -*-
#!/usr/bin/env python
"""
======================================================
NanoGen GUI CLI (:mod:`sknano.scripts.nanogenui`)
======================================================

.. currentmodule:: sknano.scripts.nanogenui

"""
from __future__ import absolute_import, print_function, division
__docformat__ = 'restructuredtext en'

import sys

from ..nanogen_gui import NGController, NGModel

__all__ = ['NanoGen']


class NanoGen(object):
    """Base class for instantiating the NanoGen MVC.

    .. versionadded:: 0.2.24

    .. seealso:: CLI module :py:mod:`sknano.scripts.nanogen`

    """
    def __init__(self, args):
        model = NGModel()
        NGController(args, model)


def main():
    NanoGen(sys.argv)

if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
"""
=============================================================================
Generate SWNT junctions (:mod:`sknano.nanogen._nanotube_junction_generators`)
=============================================================================

.. currentmodule:: sknano.nanogen._nanotube_junction_generators

"""
from __future__ import absolute_import, division, print_function
__docformat__ = 'restructuredtext en'

#import copy
#import itertools
#import sys

#import numpy as np

#from ..chemistry import Atom, Atoms
#from ..tools import plural_word_check
from ..tools.refdata import CCbond

from ._nanotube_generators import NanotubeGenerator

__all__ = ['NanotubeJunctionGenerator']


class NanotubeJunctionGenerator(NanotubeGenerator):
    u"""Class for generating nanotube junction structures.

    Parameters
    ----------
    n, m : int
        Chiral indices defining the nanotube chiral vector
        :math:`\\mathbf{C}_{h} = n\\mathbf{a}_{1} + m\\mathbf{a}_{2} = (n, m)`.
    nz : int, optional
        Number of repeat unit cells in the :math:`z` direction, along
        the *length* of the nanotube.
    element1, element2 : {str, int}, optional
        Element symbol or atomic number of basis
        :py:class:`~sknano.chemistry.Atoms` 1 and 2
    bond : float, optional
        :math:`\\mathrm{a}_{\\mathrm{CC}} =` distance between
        nearest neighbor atoms. Must be in units of **Angstroms**.
    Lz : float, optional
        Length of nanotube in units of **nanometers**.
        Overrides the `nz` value.
    fix_Lz : bool, optional
        Generate the nanotube with length as close to the specified
        :math:`L_z` as possible. If `True`, then
        non integer :math:`n_z` cells are permitted.
    autogen : bool, optional
        if `True`, automatically call
        :py:meth:`~NanotubeGenerator.generate_unit_cell`,
        followed by :py:meth:`~NanotubeGenerator.generate_structure_data`.
    verbose : bool, optional
        if `True`, show verbose output

    Examples
    --------

    """
    def __init__(self, n=int, m=int, nx=1, ny=1, nz=1,
                 element1='C', element2='C', bond=CCbond,
                 Lx=None, Ly=None, Lz=None, fix_Lz=False,
                 autogen=True, verbose=False):

        raise RuntimeError('This class is not yet implemented.')

        super(NanotubeJunctionGenerator, self).__init__(
            n=n, m=m, nx=nx, ny=ny, nz=nz,
            element1=element1, element2=element2,
            bond=bond, Lx=Lx, Ly=Ly, Lz=Lz, fix_Lz=fix_Lz,
            with_units=False, verbose=verbose)

        if autogen:
            super(NanotubeJunctionGenerator, self).generate_unit_cell()
            super(NanotubeJunctionGenerator, self).generate_structure_data()
